@@ -56,10 +56,10 @@
 #include "src/common/log.h"
 #include "src/common/env.h"
 #include "src/common/fd.h"
-#include "src/common/node_select.h"
 #include "src/common/macros.h"
 #include "src/common/proc_args.h"
 #include "src/common/read_config.h"
+#include "src/common/select.h"
 #include "src/common/slurm_opt.h"
 #include "src/common/slurm_protocol_api.h"
 #include "src/common/slurm_protocol_defs.h"
@@ -286,8 +286,8 @@ char *getenvp(char **env, const char *name)
 int setup_env(env_t *env, bool preserve_env)
 {
 	int rc = SLURM_SUCCESS;
-	char *addr, *dist = NULL, *lllp_dist = NULL;
-	char addrbuf[INET_ADDRSTRLEN];
+	char *addr, *dist = NULL;
+	char addrbuf[INET6_ADDRSTRLEN];
 
 	if (env == NULL)
 		return SLURM_ERROR;
@@ -326,6 +326,13 @@ int setup_env(env_t *env, bool preserve_env)
 		rc = SLURM_ERROR;
 	}
 
+	if (env->ntasks_per_gpu &&
+	    setenvf(&env->env, "SLURM_NTASKS_PER_GPU", "%d",
+		    env->ntasks_per_gpu)) {
+		error("Unable to set SLURM_NTASKS_PER_GPU");
+		rc = SLURM_ERROR;
+	}
+
 	if (env->ntasks_per_node
 	   && setenvf(&env->env, "SLURM_NTASKS_PER_NODE", "%d",
 		      env->ntasks_per_node) ) {
@@ -347,6 +354,13 @@ int setup_env(env_t *env, bool preserve_env)
 		rc = SLURM_ERROR;
 	}
 
+	if (env->ntasks_per_tres
+	    && setenvf(&env->env, "SLURM_NTASKS_PER_TRES", "%d",
+		      env->ntasks_per_tres) ) {
+		error("Unable to set SLURM_NTASKS_PER_TRES");
+		rc = SLURM_ERROR;
+	}
+
 	if (env->cpus_on_node
 	   && setenvf(&env->env, "SLURM_CPUS_ON_NODE", "%d",
 		      env->cpus_on_node) ) {
@@ -354,12 +368,14 @@ int setup_env(env_t *env, bool preserve_env)
 		rc = SLURM_ERROR;
 	}
 
-	set_distribution(env->distribution, &dist, &lllp_dist);
-	if (dist)
+	set_distribution(env->distribution, &dist);
+	if (dist) {
 		if (setenvf(&env->env, "SLURM_DISTRIBUTION", "%s", dist)) {
 			error("Can't set SLURM_DISTRIBUTION env variable");
 			rc = SLURM_ERROR;
 		}
+		xfree(dist);
+	}
 
 	if ((env->distribution & SLURM_DIST_STATE_BASE) == SLURM_DIST_PLANE)
 		if (setenvf(&env->env, "SLURM_DIST_PLANESIZE", "%u",
@@ -368,23 +384,15 @@ int setup_env(env_t *env, bool preserve_env)
 			rc = SLURM_ERROR;
 		}
 
-	if (lllp_dist)
-		if (setenvf(&env->env, "SLURM_DIST_LLLP", "%s", lllp_dist)) {
-			error("Can't set SLURM_DIST_LLLP env variable");
-			rc = SLURM_ERROR;
-		}
-
-
-	if (env->cpu_bind_type) {
+	if (env->cpu_bind_type && !env->batch_flag &&
+	    (env->stepid != SLURM_INTERACTIVE_STEP)) {
 		char *str_verbose, *str_bind1 = NULL, *str_bind2 = NULL;
 		char *str_bind_list, *str_bind_type = NULL, *str_bind = NULL;
 
-		if (!env->batch_flag) {
-			unsetenvp(env->env, "SLURM_CPU_BIND");
-			unsetenvp(env->env, "SLURM_CPU_BIND_LIST");
-			unsetenvp(env->env, "SLURM_CPU_BIND_TYPE");
-			unsetenvp(env->env, "SLURM_CPU_BIND_VERBOSE");
-		}
+		unsetenvp(env->env, "SLURM_CPU_BIND");
+		unsetenvp(env->env, "SLURM_CPU_BIND_LIST");
+		unsetenvp(env->env, "SLURM_CPU_BIND_TYPE");
+		unsetenvp(env->env, "SLURM_CPU_BIND_VERBOSE");
 
 		if (env->cpu_bind_type & CPU_BIND_VERBOSE)
 			str_verbose = "verbose";
@@ -399,8 +407,6 @@ int setup_env(env_t *env, bool preserve_env)
 			str_bind1 = "sockets";
 		} else if (env->cpu_bind_type & CPU_BIND_TO_LDOMS) {
 			str_bind1 = "ldoms";
-		} else if (env->cpu_bind_type & CPU_BIND_TO_BOARDS) {
-			str_bind1 = "boards";
 		}
 
 		if (env->cpu_bind_type & CPU_BIND_NONE) {
@@ -440,33 +446,31 @@ int setup_env(env_t *env, bool preserve_env)
 		} else
 			str_bind_type = xstrdup("");
 
-		if (!env->batch_flag) {
-			if (setenvf(&env->env, "SLURM_CPU_BIND", "%s", str_bind)) {
-				error("Unable to set SLURM_CPU_BIND");
-				rc = SLURM_ERROR;
-			}
-			if (setenvf(&env->env, "SLURM_CPU_BIND_LIST", "%s",
-				    str_bind_list)) {
-				error("Unable to set SLURM_CPU_BIND_LIST");
-				rc = SLURM_ERROR;
-			}
-			if (setenvf(&env->env, "SLURM_CPU_BIND_TYPE", "%s",
-				    str_bind_type)) {
-				error("Unable to set SLURM_CPU_BIND_TYPE");
-				rc = SLURM_ERROR;
-			}
-			if (setenvf(&env->env, "SLURM_CPU_BIND_VERBOSE", "%s",
-				    str_verbose)) {
-				error("Unable to set SLURM_CPU_BIND_VERBOSE");
-				rc = SLURM_ERROR;
-			}
+		if (setenvf(&env->env, "SLURM_CPU_BIND", "%s", str_bind)) {
+			error("Unable to set SLURM_CPU_BIND");
+			rc = SLURM_ERROR;
+		}
+		if (setenvf(&env->env, "SLURM_CPU_BIND_LIST", "%s",
+			    str_bind_list)) {
+			error("Unable to set SLURM_CPU_BIND_LIST");
+			rc = SLURM_ERROR;
+		}
+		if (setenvf(&env->env, "SLURM_CPU_BIND_TYPE", "%s",
+			    str_bind_type)) {
+			error("Unable to set SLURM_CPU_BIND_TYPE");
+			rc = SLURM_ERROR;
+		}
+		if (setenvf(&env->env, "SLURM_CPU_BIND_VERBOSE", "%s",
+			    str_verbose)) {
+			error("Unable to set SLURM_CPU_BIND_VERBOSE");
+			rc = SLURM_ERROR;
 		}
 
 		xfree(str_bind);
 		xfree(str_bind_type);
 	}
 
-	if (env->mem_bind_type) {
+	if (env->mem_bind_type && (env->stepid != SLURM_INTERACTIVE_STEP)) {
 		char *str_verbose, *str_bind_type = NULL, *str_bind_list;
 		char *str_prefer = NULL, *str_bind = NULL;
 		char *str_bind_sort = NULL;
@@ -707,6 +711,13 @@ int setup_env(env_t *env, bool preserve_env)
 		rc = SLURM_ERROR;
 	}
 
+	if (!preserve_env && env->threads_per_core &&
+	    setenvf(&env->env, "SLURM_THREADS_PER_CORE", "%d",
+		    env->threads_per_core)) {
+		error("Can't set SLURM_THREADS_PER_CORE env variable");
+		rc = SLURM_ERROR;
+	}
+
 	if (env->comm_port
 	    && setenvf (&env->env, "SLURM_SRUN_COMM_PORT", "%u",
 			env->comm_port)) {
@@ -715,17 +726,8 @@ int setup_env(env_t *env, bool preserve_env)
 	}
 
 	if (env->cli) {
-
-		slurm_print_slurm_addr (env->cli, addrbuf, INET_ADDRSTRLEN);
-
-		/*
-		 *  XXX: Eventually, need a function for slurm_addrs that
-		 *   returns just the IP address (not addr:port)
-		 */
-
-		if ((dist = strchr (addrbuf, ':')) != NULL)
-			*dist = '\0';
-		setenvf (&env->env, "SLURM_LAUNCH_NODE_IPADDR", "%s", addrbuf);
+		slurm_get_ip_str(env->cli, addrbuf, INET6_ADDRSTRLEN);
+		setenvf(&env->env, "SLURM_LAUNCH_NODE_IPADDR", "%s", addrbuf);
 	}
 
 	if (env->sgtids &&
@@ -799,14 +801,13 @@ int setup_env(env_t *env, bool preserve_env)
 		}
 	}
 
-	if (slurmctld_conf.slurmctld_addr)
-		addr = slurmctld_conf.slurmctld_addr;
+	if (slurm_conf.slurmctld_addr)
+		addr = slurm_conf.slurmctld_addr;
 	else
-		addr = slurmctld_conf.control_addr[0];
+		addr = slurm_conf.control_addr[0];
 	setenvf(&env->env, "SLURM_WORKING_CLUSTER", "%s:%s:%d:%d:%d",
-		slurmctld_conf.cluster_name, addr,
-		slurmctld_conf.slurmctld_port, SLURM_PROTOCOL_VERSION,
-		select_get_plugin_id());
+	        slurm_conf.cluster_name, addr, slurm_conf.slurmctld_port,
+	        SLURM_PROTOCOL_VERSION, select_get_plugin_id());
 
 	return rc;
 }
@@ -928,7 +929,7 @@ extern int env_array_for_job(char ***dest,
 			     const job_desc_msg_t *desc, int het_job_offset)
 {
 	char *tmp = NULL;
-	char *dist = NULL, *lllp_dist = NULL;
+	char *dist = NULL;
 	char *key, *value;
 	slurm_step_layout_t *step_layout = NULL;
 	int i, rc = SLURM_SUCCESS;
@@ -962,20 +963,17 @@ extern int env_array_for_job(char ***dest,
 	env_array_overwrite_het_fmt(dest, "SLURM_JOB_PARTITION", het_job_offset,
 				    "%s", alloc->partition);
 
-	set_distribution(desc->task_dist, &dist, &lllp_dist);
+	set_distribution(desc->task_dist, &dist);
 	if (dist) {
 		env_array_overwrite_het_fmt(dest, "SLURM_DISTRIBUTION",
 					    het_job_offset, "%s", dist);
+		xfree(dist);
 	}
 	if ((desc->task_dist & SLURM_DIST_STATE_BASE) == SLURM_DIST_PLANE) {
 		env_array_overwrite_het_fmt(dest, "SLURM_DIST_PLANESIZE",
 					    het_job_offset, "%u",
 					    desc->plane_size);
 	}
-	if (lllp_dist) {
-		env_array_overwrite_het_fmt(dest, "SLURM_DIST_LLLP",
-					    het_job_offset, "%s", lllp_dist);
-	}
 	tmp = uint32_compressed_to_str(alloc->num_cpu_groups,
 					alloc->cpus_per_node,
 					alloc->cpu_count_reps);
@@ -983,6 +981,11 @@ extern int env_array_for_job(char ***dest,
 				    het_job_offset, "%s", tmp);
 	xfree(tmp);
 
+	if (desc->threads_per_core != NO_VAL16)
+		env_array_overwrite_het_fmt(dest, "SLURM_THREADS_PER_CORE",
+					    het_job_offset, "%d",
+					    desc->threads_per_core);
+
 	if (alloc->pn_min_memory & MEM_PER_CPU) {
 		uint64_t tmp_mem = alloc->pn_min_memory & (~MEM_PER_CPU);
 		env_array_overwrite_het_fmt(dest, "SLURM_MEM_PER_CPU",
@@ -1003,23 +1006,32 @@ extern int env_array_for_job(char ***dest,
 	env_array_overwrite_het_fmt(dest, "SLURM_NODELIST", het_job_offset, "%s",
 				    alloc->node_list);
 
+	/*
+	 * If we know how many tasks we are going to do then we set
+	 * SLURM_TASKS_PER_NODE. If no tasks were given we can figure it out
+	 * here by totalling up the number of tasks each node can hold (which is
+	 * the cpus in a node divided by the number of cpus per task).
+	 */
 	if (step_layout_req.num_tasks == NO_VAL) {
-		/* If we know how many tasks we are going to do then
-		   we set SLURM_TASKS_PER_NODE */
-		int i = 0;
-		/* If no tasks were given we can figure it out here
-		 * by totalling up the cpus and then dividing by the
-		 * number of cpus per task */
-
 		step_layout_req.num_tasks = 0;
-		for (i = 0; i < alloc->num_cpu_groups; i++) {
-			step_layout_req.num_tasks += alloc->cpu_count_reps[i]
-				* alloc->cpus_per_node[i];
+
+		/* Iterate over all kind of cluster nodes. */
+		for (int i = 0; i < alloc->num_cpu_groups; i++) {
+			/* Get the CPU count for this type of nodes. */
+			uint32_t ntasks = alloc->cpus_per_node[i];
+
+			/*
+			 * If CPUs/tasks is set, determine how many tasks a node
+			 * of this type can hold.
+			 */
+			if ((desc->cpus_per_task != NO_VAL16) &&
+			    (desc->cpus_per_task > 1))
+				ntasks /= desc->cpus_per_task;
+
+			/* Accum. the number of tasks all the group can hold. */
+			step_layout_req.num_tasks += ntasks *
+						     alloc->cpu_count_reps[i];
 		}
-		if ((int)desc->cpus_per_task > 1
-		   && desc->cpus_per_task != NO_VAL16)
-			step_layout_req.num_tasks /= desc->cpus_per_task;
-		//num_tasks = desc->min_cpus;
 	}
 
 	if ((desc->task_dist & SLURM_DIST_STATE_BASE) == SLURM_DIST_ARBITRARY) {
@@ -1150,8 +1162,7 @@ extern int
 env_array_for_batch_job(char ***dest, const batch_job_launch_msg_t *batch,
 			const char *node_name)
 {
-	char *tmp = NULL, *cluster_name;
-	uint32_t num_cpus = 0;
+	char *tmp = NULL;
 	int i;
 	slurm_step_layout_t *step_layout = NULL;
 	uint16_t cpus_per_task;
@@ -1172,15 +1183,10 @@ env_array_for_batch_job(char ***dest, const batch_job_launch_msg_t *batch,
 	 */
 	for (i = 0; i < batch->num_cpu_groups; i++) {
 		step_layout_req.num_hosts += batch->cpu_count_reps[i];
-		num_cpus += batch->cpu_count_reps[i] * batch->cpus_per_node[i];
 	}
 
-	cluster_name = slurm_get_cluster_name();
-	if (cluster_name) {
-		env_array_overwrite_fmt(dest, "SLURM_CLUSTER_NAME", "%s",
-					cluster_name);
-		xfree(cluster_name);
-	}
+	env_array_overwrite_fmt(dest, "SLURM_CLUSTER_NAME", "%s",
+	                        slurm_conf.cluster_name);
 
 	env_array_overwrite_fmt(dest, "SLURM_JOB_ID", "%u", batch->job_id);
 	env_array_overwrite_fmt(dest, "SLURM_JOB_NUM_NODES", "%u",
@@ -1234,7 +1240,14 @@ env_array_for_batch_job(char ***dest, const batch_job_launch_msg_t *batch,
 		env_array_append_fmt(dest, "SLURM_NPROCS", "%u",
 				     step_layout_req.num_tasks);
 	} else {
-		step_layout_req.num_tasks = num_cpus / cpus_per_task;
+		/*
+		 * Iterate over all kind of cluster nodes, and accum. the number
+		 * of tasks all the group can hold.
+		 */
+		for (int i = 0; i < batch->num_cpu_groups; i++)
+			step_layout_req.num_tasks += (batch->cpus_per_node[i] /
+						      cpus_per_task) *
+						     batch->cpu_count_reps[i];
 	}
 
 	if ((step_layout_req.node_list =
@@ -1691,7 +1704,7 @@ void env_array_set_environment(char **env_array)
  * Unset all of the environment variables in a user's current
  * environment.
  *
- * (Note: becuae the environ array is decrementing with each
+ * (Note: because the environ array is decrementing with each
  *  unsetenv, only increment the ptr on a failure to unset.)
  */
 void env_unset_environment(void)
@@ -1884,22 +1897,50 @@ char **env_array_from_file(const char *fname)
 	return env;
 }
 
+int env_array_to_file(const char *filename, const char **env_array)
+{
+	int outfd = -1;
+	int rc = SLURM_SUCCESS;
+
+	outfd = open(filename, (O_WRONLY | O_CREAT | O_EXCL), 0600);
+	if (outfd < 0) {
+		error("%s: unable to open %s: %m",
+		      __func__, filename);
+		goto rwfail;
+	}
+
+	for (const char **p = env_array; p && *p; p++) {
+		safe_write(outfd, *p, strlen(*p));
+		safe_write(outfd, "\0", 1);
+	}
+
+	(void) close(outfd);
+
+	return rc;
+
+rwfail:
+	rc = errno;
+
+	if (outfd >= 0)
+		(void) close(outfd);
+
+	return rc;
+}
+
 /*
  * Load user environment from a cache file located in
  * <state_save_location>/env_username
  */
 static char **_load_env_cache(const char *username)
 {
-	char *state_save_loc, fname[MAXPATHLEN];
+	char fname[MAXPATHLEN];
 	char *line, name[256], *value;
 	char **env = NULL;
 	FILE *fp;
 	int i;
 
-	state_save_loc = slurm_get_state_save_location();
-	i = snprintf(fname, sizeof(fname), "%s/env_cache/%s", state_save_loc,
-		     username);
-	xfree(state_save_loc);
+	i = snprintf(fname, sizeof(fname), "%s/env_cache/%s",
+		     slurm_conf.state_save_location, username);
 	if (i < 0) {
 		error("Environment cache filename overflow");
 		return NULL;
@@ -1971,7 +2012,7 @@ char **env_array_user_default(const char *username, int timeout, int mode,
 	char *stoptoken  = "XXXXSLURMSTOPPARSINGHEREXXXXX";
 	char cmdstr[256], *env_loc = NULL;
 	char *stepd_path = NULL;
-	int fd1, fd2, fildes[2], found, fval, len, rc, timeleft;
+	int fildes[2], found, fval, len, rc, timeleft;
 	int buf_read, buf_rem, config_timeout;
 	pid_t child;
 	struct timeval begin, now;
@@ -1983,9 +2024,7 @@ char **env_array_user_default(const char *username, int timeout, int mode,
 		return NULL;
 	}
 
-	config_timeout = slurm_get_env_timeout();
-
-	if (config_timeout == 0)	/* just read directly from cache */
+	if (!slurm_conf.get_env_timeout)	/* just read directly from cache */
 		return _load_env_cache(username);
 
 	if (stat(SUCMD, &buf))
@@ -2001,7 +2040,7 @@ char **env_array_user_default(const char *username, int timeout, int mode,
 	else if (stat("/usr/bin/env", &buf) == 0)
 		env_loc = "/usr/bin/env";
 	else
-		fatal("Could not location command: env");
+		fatal("Could not locate command: env");
 	snprintf(cmdstr, sizeof(cmdstr),
 		 "/bin/echo; /bin/echo; /bin/echo; "
 		 "/bin/echo %s; %s; /bin/echo %s",
@@ -2019,31 +2058,33 @@ char **env_array_user_default(const char *username, int timeout, int mode,
 		return NULL;
 	}
 	if (child == 0) {
-		setenv("ENVIRONMENT", "BATCH", 1);
+		char **tmp_env = NULL;
+		int devnull;
+		tmp_env = env_array_create();
+		env_array_overwrite(&tmp_env, "ENVIRONMENT", "BATCH");
 		setpgid(0, 0);
-		close(0);
-		if ((fd1 = open("/dev/null", O_RDONLY)) == -1)
-			error("%s: open(/dev/null): %m", __func__);
-		dup2(fildes[1], 1);
-		close(2);
-		if ((fd2 = open("/dev/null", O_WRONLY)) == -1)
+
+		if ((devnull = open("/dev/null", O_RDONLY)) == -1)
 			error("%s: open(/dev/null): %m", __func__);
+		dup2(devnull, STDIN_FILENO);
+		dup2(fildes[1], STDOUT_FILENO);
+		dup2(devnull, STDERR_FILENO);
+		closeall(3);
+
 		if      (mode == 1)
-			execl(SUCMD, "su", username, "-c", cmdstr, NULL);
+			execle(SUCMD, "su", username, "-c", cmdstr, NULL, tmp_env);
 		else if (mode == 2)
-			execl(SUCMD, "su", "-", username, "-c", cmdstr, NULL);
+			execle(SUCMD, "su", "-", username, "-c", cmdstr, NULL, tmp_env);
 		else {	/* Default system configuration */
 #ifdef LOAD_ENV_NO_LOGIN
-			execl(SUCMD, "su", username, "-c", cmdstr, NULL);
+			execle(SUCMD, "su", username, "-c", cmdstr, NULL, tmp_env);
 #else
-			execl(SUCMD, "su", "-", username, "-c", cmdstr, NULL);
+			execle(SUCMD, "su", "-", username, "-c", cmdstr, NULL, tmp_env);
 #endif
 		}
-		if (fd1 >= 0)	/* Avoid Coverity resource leak notification */
-			(void) close(fd1);
-		if (fd2 >= 0)	/* Avoid Coverity resource leak notification */
-			(void) close(fd2);
-		exit(1);
+		if (devnull >= 0)	/* Avoid Coverity resource leak notification */
+			(void) close(devnull);
+		_exit(1);
 	}
 
 	close(fildes[1]);
@@ -2058,7 +2099,7 @@ char **env_array_user_default(const char *username, int timeout, int mode,
 
 	/* Read all of the output from /bin/su into buffer */
 	if (timeout == 0)
-		timeout = config_timeout;	/* != 0 test above */
+		timeout = slurm_conf.get_env_timeout;	/* != 0 test above */
 	found = 0;
 	buf_read = 0;
 	buffer = xmalloc(ENV_BUFSIZE);
@@ -2291,3 +2332,49 @@ extern char *find_quote_token(char *tmp, char *sep, char **last)
 
 	}
 }
+
+extern void env_merge_filter(slurm_opt_t *opt, job_desc_msg_t *desc)
+{
+	extern char **environ;
+	int i, len;
+	char *save_env[2] = { NULL, NULL }, *tmp, *tok, *last = NULL;
+
+	tmp = xstrdup(opt->export_env);
+	tok = find_quote_token(tmp, ",", &last);
+	while (tok) {
+
+		if (xstrcasecmp(tok, "ALL") == 0) {
+			env_array_merge(&desc->environment,
+					(const char **)environ);
+			tok = find_quote_token(NULL, ",", &last);
+			continue;
+		}
+
+		if (strchr(tok, '=')) {
+			save_env[0] = tok;
+			env_array_merge(&desc->environment,
+					(const char **)save_env);
+		} else {
+			len = strlen(tok);
+			for (i = 0; environ[i]; i++) {
+				if (xstrncmp(tok, environ[i], len) ||
+				    (environ[i][len] != '='))
+					continue;
+				save_env[0] = environ[i];
+				env_array_merge(&desc->environment,
+						(const char **)save_env);
+				break;
+			}
+		}
+		tok = find_quote_token(NULL, ",", &last);
+	}
+	xfree(tmp);
+
+	for (i = 0; environ[i]; i++) {
+		if (xstrncmp("SLURM_", environ[i], 6))
+			continue;
+		save_env[0] = environ[i];
+		env_array_merge(&desc->environment,
+				(const char **)save_env);
+	}
+}
