@@ -3,31 +3,31 @@
  *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
  *  Written by Chris Dunlap <cdunlap@llnl.gov>.
  *  UCRL-CODE-2002-009.
- *  
+ *
  *  This file is part of ConMan, a remote console management program.
  *  For details, see <http://www.llnl.gov/linux/conman/>.
- *  
+ *
  *  ConMan is free software; you can redistribute it and/or modify it under
  *  the terms of the GNU General Public License as published by the Free
  *  Software Foundation; either version 2 of the License, or (at your option)
  *  any later version.
  *
- *  In addition, as a special exception, the copyright holders give permission 
- *  to link the code of portions of this program with the OpenSSL library under 
- *  certain conditions as described in each individual source file, and 
- *  distribute linked combinations including the two. You must obey the GNU 
- *  General Public License in all respects for all of the code used other than 
- *  OpenSSL. If you modify file(s) with this exception, you may extend this 
- *  exception to your version of the file(s), but you are not obligated to do 
+ *  In addition, as a special exception, the copyright holders give permission
+ *  to link the code of portions of this program with the OpenSSL library under
+ *  certain conditions as described in each individual source file, and
+ *  distribute linked combinations including the two. You must obey the GNU
+ *  General Public License in all respects for all of the code used other than
+ *  OpenSSL. If you modify file(s) with this exception, you may extend this
+ *  exception to your version of the file(s), but you are not obligated to do
  *  so. If you do not wish to do so, delete this exception statement from your
- *  version.  If you delete this exception statement from all source files in 
+ *  version.  If you delete this exception statement from all source files in
  *  the program, then also delete it here.
- *  
+ *
  *  ConMan is distributed in the hope that it will be useful, but WITHOUT ANY
  *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  *  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  *  details.
- *  
+ *
  *  You should have received a copy of the GNU General Public License along
  *  with ConMan; if not, write to the Free Software Foundation, Inc.,
  *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA.
@@ -35,6 +35,7 @@
  *  Refer to "fd.h" for documentation on public functions.
 \*****************************************************************************/
 
+#include <dirent.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <inttypes.h>
@@ -48,24 +49,71 @@
 #include "src/common/fd.h"
 #include "src/common/log.h"
 #include "src/common/macros.h"
+#include "src/common/net.h"
+#include "src/common/read_config.h"
+#include "src/common/slurm_protocol_api.h"
 #include "src/common/timers.h"
 #include "src/common/xassert.h"
 #include "src/common/xmalloc.h"
 #include "src/common/xstring.h"
 
 /*
- * Define slurm-specific aliases for use by plugins, see slurm_xlator.h 
- * for details. 
+ * Define slurm-specific aliases for use by plugins, see slurm_xlator.h
+ * for details.
  */
+strong_alias(closeall, slurm_closeall);
 strong_alias(fd_set_blocking,	slurm_fd_set_blocking);
 strong_alias(fd_set_nonblocking,slurm_fd_set_nonblocking);
 strong_alias(fd_get_socket_error, slurm_fd_get_socket_error);
 strong_alias(send_fd_over_pipe, slurm_send_fd_over_pipe);
 strong_alias(receive_fd_over_pipe, slurm_receive_fd_over_pipe);
+strong_alias(rmdir_recursive, slurm_rmdir_recursive);
 
 static int fd_get_lock(int fd, int cmd, int type);
 static pid_t fd_test_lock(int fd, int type);
 
+static void _slow_closeall(int fd)
+{
+	struct rlimit rlim;
+
+	if (getrlimit(RLIMIT_NOFILE, &rlim) < 0) {
+		error("getrlimit(RLIMIT_NOFILE): %m");
+		rlim.rlim_cur = 4096;
+	}
+
+	while (fd < rlim.rlim_cur)
+		close(fd++);
+}
+
+extern void closeall(int fd)
+{
+	char *name = "/proc/self/fd";
+	DIR *d;
+	struct dirent *dir;
+
+	/*
+	 * Blindly closing all file descriptors is slow.
+	 *
+	 * Instead get all open file descriptors from /proc/self/fd, then
+	 * close each one of those that are greater than or equal to fd.
+	 */
+	if (!(d = opendir(name))) {
+		debug("Could not read open files from %s: %m, closing all potential file descriptors",
+		      name);
+		_slow_closeall(fd);
+		return;
+	}
+
+	while ((dir = readdir(d))) {
+		/* Ignore "." and ".." entries */
+		if (dir->d_type != DT_DIR) {
+			int open_fd = atoi(dir->d_name);
+			if (open_fd >= fd)
+				close(open_fd);
+		}
+	}
+	closedir(d);
+}
 
 void fd_set_close_on_exec(int fd)
 {
@@ -139,10 +187,20 @@ int fd_get_socket_error(int fd, int *err)
 
 	xassert(fd >= 0);
 
+	*err = SLURM_SUCCESS;
+
 	if (getsockopt(fd, SOL_SOCKET, SO_ERROR, (void *)err, &errlen))
 		return errno;
-	else
+	else {
+		/*
+		 * SOL_SOCKET/SO_ERROR may not find an error and will not set
+		 * errno. This may happen if on duplicate calls or if something
+		 * else has cleared the error.
+		 */
+		if (!(*err))
+			*err = SLURM_COMMUNICATIONS_MISSING_SOCKET_ERROR;
 		return SLURM_SUCCESS;
+	}
 }
 
 static int fd_get_lock(int fd, int cmd, int type)
@@ -283,6 +341,28 @@ extern char *fd_resolve_path(int fd)
 	return resolved;
 }
 
+extern char *fd_resolve_peer(int fd)
+{
+	slurm_addr_t addr;
+	socklen_t size = sizeof(addr);
+	int err = errno;
+	char *peer;
+
+	if (fd < 0)
+		return NULL;
+
+	if (slurm_get_peer_addr(fd, &addr)) {
+		log_flag(NET, "%s: unable to resolve peername for fd:%d: %m",
+			 __func__, fd);
+		return NULL;
+	}
+
+	peer = sockaddr_to_string(&addr, size);
+
+	errno = err;
+	return peer;
+}
+
 extern void fd_set_oob(int fd, int value)
 {
 	if (setsockopt(fd, SOL_SOCKET, SO_OOBINLINE, &value, sizeof(value)))
@@ -309,7 +389,7 @@ extern char *poll_revents_to_str(const short revents)
 	if (!revents)
 		xstrfmtcat(txt, "0");
 	else
-		xstrfmtcat(txt, "(0x%04" PRIx16 ")", revents);
+		xstrfmtcat(txt, "(0x%04x)", revents);
 
 	return txt;
 }
@@ -368,10 +448,99 @@ extern int receive_fd_over_pipe(int socket)
 
 	cmsg = CMSG_FIRSTHDR(&msg);
 	if (!cmsg) {
-		error("%s: CMSG_FIRSTHDR error: %m", __func__);
+		error("%s: CMSG_FIRSTHDR failed", __func__);
 		return -1;
 	}
 	memmove(&fd, CMSG_DATA(cmsg), sizeof(fd));
 
 	return fd;
 }
+
+static int _rmdir_recursive(int dirfd)
+{
+	int rc = 0;
+	DIR *dp;
+	struct dirent *ent;
+
+	if (!(dp = fdopendir(dirfd))) {
+		error("%s: can't open directory: %m", __func__);
+		return 1;
+	}
+
+	while ((ent = readdir(dp))) {
+		int childfd = -1;
+
+		/* skip special directories */
+		if (!strcmp(ent->d_name, ".") ||
+		    !strcmp(ent->d_name, "..")) {
+			continue;
+		}
+
+		/* try to remove entry, first as a file, then as a directory */
+		if (unlinkat(dirfd, ent->d_name, 0) != -1) {
+			debug("%s: removed file `%s`", __func__, ent->d_name);
+			continue;
+		} else if (unlinkat(dirfd, ent->d_name, AT_REMOVEDIR) != -1) {
+			debug("%s: removed empty directory `%s`",
+			      __func__, ent->d_name);
+			continue;
+		}
+
+		/* removal didn't work. assume it's a non-empty directory */
+		if ((childfd = openat(dirfd, ent->d_name,
+				      (O_DIRECTORY | O_NOFOLLOW))) < 0) {
+			debug("%s: openat() failed for `%s`: %m",
+			      __func__, ent->d_name);
+			rc++;
+			continue;
+		}
+
+		debug("%s: descending into directory `%s`",
+		      __func__, ent->d_name);
+		rc += _rmdir_recursive(childfd);
+		(void) close(childfd);
+
+		if (unlinkat(dirfd, ent->d_name, AT_REMOVEDIR) != -1) {
+			debug("%s: removed now-empty directory `%s`",
+			      __func__, ent->d_name);
+		} else {
+			debug("%s: unlinkat() failed for `%s`: %m",
+			      __func__, ent->d_name);
+			rc++;
+		}
+	}
+	closedir(dp);
+
+	return rc;
+}
+
+extern int rmdir_recursive(const char *path, bool remove_top)
+{
+	int rc = 0;
+	int dirfd;
+
+	if ((dirfd = open(path, O_DIRECTORY | O_NOFOLLOW)) < 0) {
+		error("%s: could not open %s", __func__, path);
+		return 1;
+	}
+
+	rc = _rmdir_recursive(dirfd);
+	close(dirfd);
+
+	if (remove_top) {
+		if (rmdir(path) < 0) {
+			debug("%s: rmdir() failed for `%s`: %m",
+			      __func__, path);
+			rc++;
+		} else {
+			debug("%s: removed now-empty top directory `%s`",
+			      __func__, path);
+		}
+	}
+
+	if (rc)
+		error("%s: could not completely remove `%s`, %d files left",
+		      __func__, path, rc);
+
+	return rc;
+}
