@@ -68,6 +68,7 @@
 #include "src/common/read_config.h"
 #include "src/common/slurm_acct_gather_profile.h"
 #include "src/common/slurm_jobacct_gather.h"
+#include "src/common/slurm_protocol_pack.h"
 #include "src/common/slurmdbd_defs.h"
 #include "src/common/xmalloc.h"
 #include "src/common/xstring.h"
@@ -87,8 +88,7 @@ strong_alias(jobacctinfo_create, slurm_jobacctinfo_create);
 strong_alias(jobacctinfo_destroy, slurm_jobacctinfo_destroy);
 
 typedef struct slurm_jobacct_gather_ops {
-	void (*poll_data) (List task_list, bool pgid_plugin, uint64_t cont_id,
-			   bool profile);
+	void (*poll_data) (List task_list, uint64_t cont_id, bool profile);
 	int (*endpoll)    ();
 	int (*add_task)   (pid_t pid, jobacct_id_t *jobacct_id);
 } slurm_jobacct_gather_ops_t;
@@ -111,7 +111,6 @@ static pthread_mutex_t init_run_mutex = PTHREAD_MUTEX_INITIALIZER;
 static pthread_t watch_tasks_thread_id = 0;
 
 static int freq = 0;
-static bool pgid_plugin = false;
 static List task_list = NULL;
 static uint64_t cont_id = NO_VAL64;
 static pthread_mutex_t task_list_lock = PTHREAD_MUTEX_INITIALIZER;
@@ -120,8 +119,11 @@ static bool jobacct_shutdown = true;
 static pthread_mutex_t jobacct_shutdown_mutex = PTHREAD_MUTEX_INITIALIZER;
 static bool plugin_polling = true;
 
-static uint32_t jobacct_job_id     = 0;
-static uint32_t jobacct_step_id    = 0;
+static slurm_step_id_t jobacct_step_id = {
+	.job_id = 0,
+	.step_het_comp = NO_VAL,
+	.step_id = 0,
+};
 static uint64_t jobacct_mem_limit  = 0;
 static uint64_t jobacct_vmem_limit = 0;
 static acct_gather_profile_timer_t *profile_timer =
@@ -284,8 +286,8 @@ static void _acct_kill_step(void)
 	job_notify_msg_t notify_req;
 
 	slurm_msg_t_init(&msg);
-	notify_req.job_id      = jobacct_job_id;
-	notify_req.job_step_id = jobacct_step_id;
+	memcpy(&notify_req.step_id, &jobacct_step_id,
+	       sizeof(notify_req.step_id));
 	notify_req.message     = "Exceeded job memory limit";
 	msg.msg_type    = REQUEST_JOB_NOTIFY;
 	msg.data        = &notify_req;
@@ -295,8 +297,7 @@ static void _acct_kill_step(void)
 	 * Request message:
 	 */
 	memset(&req, 0, sizeof(job_step_kill_msg_t));
-	req.job_id      = jobacct_job_id;
-	req.job_step_id = jobacct_step_id;
+	memcpy(&req.step_id, &jobacct_step_id, sizeof(req.step_id));
 	req.signal      = SIGKILL;
 	req.flags       = 0;
 	msg.msg_type    = REQUEST_CANCEL_JOB_STEP;
@@ -319,7 +320,7 @@ static void _poll_data(bool profile)
 	/* Update the data */
 	slurm_mutex_lock(&task_list_lock);
 	if (task_list)
-		(*(ops.poll_data))(task_list, pgid_plugin, cont_id, profile);
+		(*(ops.poll_data))(task_list, cont_id, profile);
 	slurm_mutex_unlock(&task_list_lock);
 }
 
@@ -379,6 +380,8 @@ static void _jobacctinfo_aggregate_tres_usage(jobacctinfo_t *dest,
 {
 	uint32_t i = 0;
 
+	xassert(dest->tres_count == from->tres_count);
+
 	for (i = 0; i < dest->tres_count; i++) {
 		if (from->tres_usage_in_max[i] != INFINITE64) {
 			if ((dest->tres_usage_in_max[i] == INFINITE64) ||
@@ -519,7 +522,6 @@ static void _jobacctinfo_2_stats_tres_usage(slurmdb_stats_t *stats,
 extern int jobacct_gather_init(void)
 {
 	char    *plugin_type = "jobacct_gather";
-	char	*type = NULL;
 	int	retval=SLURM_SUCCESS;
 
 	if (slurmdbd_conf || (_init_run_test() && g_context))
@@ -529,18 +531,19 @@ extern int jobacct_gather_init(void)
 	if (g_context)
 		goto done;
 
-	type = slurm_get_jobacct_gather_type();
-
-	g_context = plugin_context_create(
-		plugin_type, type, (void **)&ops, syms, sizeof(syms));
+	g_context = plugin_context_create(plugin_type,
+					  slurm_conf.job_acct_gather_type,
+					  (void **) &ops, syms, sizeof(syms));
 
 	if (!g_context) {
-		error("cannot create %s context for %s", plugin_type, type);
+		error("cannot create %s context for %s",
+		      plugin_type, slurm_conf.job_acct_gather_type);
 		retval = SLURM_ERROR;
 		goto done;
 	}
 
-	if (!xstrcasecmp(type, "jobacct_gather/none")) {
+	if (!xstrcasecmp(slurm_conf.job_acct_gather_type,
+			 "jobacct_gather/none")) {
 		plugin_polling = false;
 		goto done;
 	}
@@ -553,20 +556,12 @@ extern int jobacct_gather_init(void)
 	if (!running_in_slurmctld())
 		goto done;
 
-	plugin_type = type;
-	type = slurm_get_proctrack_type();
-	if (!xstrcasecmp(type, "proctrack/pgid")) {
-		info("WARNING: We will use a much slower algorithm with "
-		     "proctrack/pgid, use Proctracktype=proctrack/linuxproc "
-		     "or some other proctrack when using %s",
-		     plugin_type);
-		pgid_plugin = true;
-	}
-	xfree(type);
-	xfree(plugin_type);
+	if (!xstrcasecmp(slurm_conf.proctrack_type, "proctrack/pgid"))
+		info("WARNING: We will use a much slower algorithm with proctrack/pgid, use Proctracktype=proctrack/linuxproc or some other proctrack when using %s",
+		     slurm_conf.job_acct_gather_type);
 
-	type = slurm_get_accounting_storage_type();
-	if (!xstrcasecmp(type, ACCOUNTING_STORAGE_TYPE_NONE)) {
+	if (!xstrcasecmp(slurm_conf.accounting_storage_type,
+	                 ACCOUNTING_STORAGE_TYPE_NONE)) {
 		error("WARNING: Even though we are collecting accounting "
 		      "information you have asked for it not to be stored "
 		      "(%s) if this is not what you have in mind you will "
@@ -575,7 +570,6 @@ extern int jobacct_gather_init(void)
 
 done:
 	slurm_mutex_unlock(&g_context_lock);
-	xfree(type);
 
 	return(retval);
 }
@@ -690,11 +684,10 @@ extern int jobacct_gather_add_task(pid_t pid, jobacct_id_t *jobacct_id,
 	memcpy(&jobacct->id, jobacct_id, sizeof(jobacct_id_t));
 	debug2("adding task %u pid %d on node %u to jobacct",
 	       jobacct_id->taskid, pid, jobacct_id->nodeid);
+	(*(ops.add_task))(pid, jobacct_id);
 	list_push(task_list, jobacct);
 	slurm_mutex_unlock(&task_list_lock);
 
-	(*(ops.add_task))(pid, jobacct_id);
-
 	if (poll == 1)
 		_poll_data(1);
 
@@ -765,7 +758,7 @@ extern jobacctinfo_t *jobacct_gather_remove_task(pid_t pid)
 
 	itr = list_iterator_create(task_list);
 	while((jobacct = list_next(itr))) {
-		if (jobacct->pid == pid) {
+		if (!pid || (jobacct->pid == pid)) {
 			list_remove(itr);
 			break;
 		}
@@ -775,7 +768,8 @@ extern jobacctinfo_t *jobacct_gather_remove_task(pid_t pid)
 		debug2("removing task %u pid %d from jobacct",
 		       jobacct->id.taskid, jobacct->pid);
 	} else {
-		debug2("pid(%d) not being watched in jobacct!", pid);
+		if (pid)
+			debug2("pid(%d) not being watched in jobacct!", pid);
 	}
 error:
 	slurm_mutex_unlock(&task_list_lock);
@@ -784,7 +778,7 @@ error:
 
 extern int jobacct_gather_set_proctrack_container_id(uint64_t id)
 {
-	if (!plugin_polling || pgid_plugin)
+	if (!plugin_polling)
 		return SLURM_SUCCESS;
 
 	if (cont_id != NO_VAL64)
@@ -802,24 +796,22 @@ extern int jobacct_gather_set_proctrack_container_id(uint64_t id)
 	return SLURM_SUCCESS;
 }
 
-extern int jobacct_gather_set_mem_limit(uint32_t job_id,
-					uint32_t step_id,
+extern int jobacct_gather_set_mem_limit(slurm_step_id_t *step_id,
 					uint64_t mem_limit)
 {
 	if (!plugin_polling)
 		return SLURM_SUCCESS;
 
-	if ((job_id == 0) || (mem_limit == 0)) {
+	if ((step_id->job_id == 0) || (mem_limit == 0)) {
 		error("jobacct_gather_set_mem_limit: jobid:%u "
-		      "mem_limit:%"PRIu64"", job_id, mem_limit);
+		      "mem_limit:%"PRIu64"", step_id->job_id, mem_limit);
 		return SLURM_ERROR;
 	}
 
-	jobacct_job_id      = job_id;
-	jobacct_step_id     = step_id;
+	memcpy(&jobacct_step_id, step_id, sizeof(jobacct_step_id));
 	jobacct_mem_limit   = mem_limit * 1048576; /* MB to B */
 	jobacct_vmem_limit  = jobacct_mem_limit;
-	jobacct_vmem_limit *= (slurm_get_vsize_factor() / 100.0);
+	jobacct_vmem_limit *= (slurm_conf.vsize_factor / 100.0);
 	return SLURM_SUCCESS;
 }
 
@@ -829,45 +821,19 @@ extern void jobacct_gather_handle_mem_limit(uint64_t total_job_mem,
 	if (!plugin_polling)
 		return;
 
-	if (jobacct_mem_limit) {
-		if (jobacct_step_id == NO_VAL) {
-			debug("Job %u memory used:%"PRIu64" limit:%"PRIu64" B",
-			      jobacct_job_id, total_job_mem, jobacct_mem_limit);
-		} else {
-			debug("Step %u.%u memory used:%"PRIu64" "
-			      "limit:%"PRIu64" B",
-			      jobacct_job_id, jobacct_step_id,
-			      total_job_mem, jobacct_mem_limit);
-		}
-	}
-	if (jobacct_job_id && jobacct_mem_limit &&
+	if (jobacct_mem_limit)
+		debug("%ps memory used:%"PRIu64" limit:%"PRIu64" B",
+		      &jobacct_step_id, total_job_mem, jobacct_mem_limit);
+
+	if (jobacct_step_id.job_id && jobacct_mem_limit &&
 	    (total_job_mem > jobacct_mem_limit)) {
-		if (jobacct_step_id == NO_VAL) {
-			error("Job %u exceeded memory limit "
-			      "(%"PRIu64" > %"PRIu64"), being "
-			      "killed", jobacct_job_id, total_job_mem,
-			      jobacct_mem_limit);
-		} else {
-			error("Step %u.%u exceeded memory limit "
-			      "(%"PRIu64" > %"PRIu64"), "
-			      "being killed", jobacct_job_id, jobacct_step_id,
-			      total_job_mem, jobacct_mem_limit);
-		}
+		error("%ps exceeded memory limit (%"PRIu64" > %"PRIu64"), being killed",
+		      &jobacct_step_id, total_job_mem, jobacct_mem_limit);
 		_acct_kill_step();
-	} else if (jobacct_job_id && jobacct_vmem_limit &&
+	} else if (jobacct_step_id.job_id && jobacct_vmem_limit &&
 		   (total_job_vsize > jobacct_vmem_limit)) {
-		if (jobacct_step_id == NO_VAL) {
-			error("Job %u exceeded virtual memory limit "
-			      "(%"PRIu64" > %"PRIu64"), being killed",
-			      jobacct_job_id,
-			      total_job_vsize, jobacct_vmem_limit);
-		} else {
-			error("Step %u.%u exceeded virtual memory limit "
-			      "(%"PRIu64" > %"PRIu64"), being killed",
-			      jobacct_job_id,
-			      jobacct_step_id, total_job_vsize,
-			      jobacct_vmem_limit);
-		}
+		error("%ps exceeded virtual memory limit (%"PRIu64" > %"PRIu64"), being killed",
+		      &jobacct_step_id, total_job_vsize, jobacct_vmem_limit);
 		_acct_kill_step();
 	}
 }
@@ -917,7 +883,7 @@ extern int jobacctinfo_setinfo(jobacctinfo_t *jobacct,
 	struct rusage *rusage = (struct rusage *)data;
 	uint64_t *uint64 = (uint64_t *) data;
 	struct jobacctinfo *send = (struct jobacctinfo *) data;
-	Buf buffer = NULL;
+	buf_t *buffer = NULL;
 
 	if (!plugin_polling)
 		return SLURM_SUCCESS;
@@ -1012,7 +978,7 @@ extern int jobacctinfo_getinfo(
 	case JOBACCT_DATA_PIPE:
 		if (protocol_version >= SLURM_MIN_PROTOCOL_VERSION) {
 			int len;
-			Buf buffer;
+			buf_t *buffer;
 
 			safe_read(*fd, &len, sizeof(int));
 			buf = xmalloc(len);
@@ -1047,9 +1013,8 @@ rwfail:
 	return SLURM_ERROR;
 }
 
-extern void jobacctinfo_pack(jobacctinfo_t *jobacct,
-			     uint16_t rpc_version, uint16_t protocol_type,
-			     Buf buffer)
+extern void jobacctinfo_pack(jobacctinfo_t *jobacct, uint16_t rpc_version,
+			     uint16_t protocol_type, buf_t *buffer)
 {
 	bool no_pack;
 
@@ -1062,10 +1027,58 @@ extern void jobacctinfo_pack(jobacctinfo_t *jobacct,
 
 	pack8((uint8_t) 1, buffer);
 
-	if (rpc_version >= SLURM_MIN_PROTOCOL_VERSION) {
-		pack32((uint32_t)jobacct->user_cpu_sec, buffer);
+	if (rpc_version >= SLURM_21_08_PROTOCOL_VERSION) {
+		pack64(jobacct->user_cpu_sec, buffer);
 		pack32((uint32_t)jobacct->user_cpu_usec, buffer);
-		pack32((uint32_t)jobacct->sys_cpu_sec, buffer);
+		pack64(jobacct->sys_cpu_sec, buffer);
+		pack32((uint32_t)jobacct->sys_cpu_usec, buffer);
+		pack32((uint32_t)jobacct->act_cpufreq, buffer);
+		pack64((uint64_t)jobacct->energy.consumed_energy, buffer);
+
+		pack32_array(jobacct->tres_ids, jobacct->tres_count, buffer);
+
+		slurm_pack_list(jobacct->tres_list,
+				slurmdb_pack_tres_rec, buffer,
+				SLURM_PROTOCOL_VERSION);
+
+		pack64_array(jobacct->tres_usage_in_max,
+			     jobacct->tres_count, buffer);
+		pack64_array(jobacct->tres_usage_in_max_nodeid,
+			     jobacct->tres_count, buffer);
+		pack64_array(jobacct->tres_usage_in_max_taskid,
+			     jobacct->tres_count, buffer);
+		pack64_array(jobacct->tres_usage_in_min,
+			     jobacct->tres_count, buffer);
+		pack64_array(jobacct->tres_usage_in_min_nodeid,
+			     jobacct->tres_count, buffer);
+		pack64_array(jobacct->tres_usage_in_min_taskid,
+			     jobacct->tres_count, buffer);
+		pack64_array(jobacct->tres_usage_in_tot,
+			     jobacct->tres_count, buffer);
+		pack64_array(jobacct->tres_usage_out_max,
+			     jobacct->tres_count, buffer);
+		pack64_array(jobacct->tres_usage_out_max_nodeid,
+			     jobacct->tres_count, buffer);
+		pack64_array(jobacct->tres_usage_out_max_taskid,
+			     jobacct->tres_count, buffer);
+		pack64_array(jobacct->tres_usage_out_min,
+			     jobacct->tres_count, buffer);
+		pack64_array(jobacct->tres_usage_out_min_nodeid,
+			     jobacct->tres_count, buffer);
+		pack64_array(jobacct->tres_usage_out_min_taskid,
+			     jobacct->tres_count, buffer);
+		pack64_array(jobacct->tres_usage_out_tot,
+			     jobacct->tres_count, buffer);
+	} else if (rpc_version >= SLURM_MIN_PROTOCOL_VERSION) {
+		if (jobacct->user_cpu_sec > NO_VAL) {
+			pack32((uint32_t)NO_VAL, buffer);
+		} else
+			pack32((uint32_t)jobacct->user_cpu_sec, buffer);
+		pack32((uint32_t)jobacct->user_cpu_usec, buffer);
+		if (jobacct->sys_cpu_sec > NO_VAL) {
+			pack32((uint32_t)NO_VAL, buffer);
+		} else
+			pack32((uint32_t)jobacct->sys_cpu_sec, buffer);
 		pack32((uint32_t)jobacct->sys_cpu_usec, buffer);
 		pack32((uint32_t)jobacct->act_cpufreq, buffer);
 		pack64((uint64_t)jobacct->energy.consumed_energy, buffer);
@@ -1110,9 +1123,8 @@ extern void jobacctinfo_pack(jobacctinfo_t *jobacct,
 	}
 }
 
-extern int jobacctinfo_unpack(jobacctinfo_t **jobacct,
-			      uint16_t rpc_version, uint16_t protocol_type,
-			      Buf buffer, bool alloc)
+extern int jobacctinfo_unpack(jobacctinfo_t **jobacct, uint16_t rpc_version,
+			      uint16_t protocol_type, buf_t *buffer, bool alloc)
 {
 	uint32_t uint32_tmp;
 	uint8_t  uint8_tmp;
@@ -1133,7 +1145,53 @@ extern int jobacctinfo_unpack(jobacctinfo_t **jobacct,
 		_free_tres_usage(*jobacct);
 	}
 
-	if (rpc_version >= SLURM_MIN_PROTOCOL_VERSION) {
+	if (rpc_version >= SLURM_21_08_PROTOCOL_VERSION) {
+		safe_unpack64(&(*jobacct)->user_cpu_sec, buffer);
+		safe_unpack32(&uint32_tmp, buffer);
+		(*jobacct)->user_cpu_usec = uint32_tmp;
+		safe_unpack64(&(*jobacct)->sys_cpu_sec, buffer);
+		safe_unpack32(&uint32_tmp, buffer);
+		(*jobacct)->sys_cpu_usec = uint32_tmp;
+
+		safe_unpack32(&(*jobacct)->act_cpufreq, buffer);
+		safe_unpack64(&(*jobacct)->energy.consumed_energy, buffer);
+
+		safe_unpack32_array(&(*jobacct)->tres_ids,
+				    &(*jobacct)->tres_count, buffer);
+		if (slurm_unpack_list(&(*jobacct)->tres_list,
+				      slurmdb_unpack_tres_rec,
+				      slurmdb_destroy_tres_rec,
+				      buffer, rpc_version) != SLURM_SUCCESS)
+			goto unpack_error;
+		safe_unpack64_array(&(*jobacct)->tres_usage_in_max,
+				    &uint32_tmp, buffer);
+		safe_unpack64_array(&(*jobacct)->tres_usage_in_max_nodeid,
+				    &uint32_tmp, buffer);
+		safe_unpack64_array(&(*jobacct)->tres_usage_in_max_taskid,
+				    &uint32_tmp, buffer);
+		safe_unpack64_array(&(*jobacct)->tres_usage_in_min,
+				    &uint32_tmp, buffer);
+		safe_unpack64_array(&(*jobacct)->tres_usage_in_min_nodeid,
+				    &uint32_tmp, buffer);
+		safe_unpack64_array(&(*jobacct)->tres_usage_in_min_taskid,
+				    &uint32_tmp, buffer);
+		safe_unpack64_array(&(*jobacct)->tres_usage_in_tot,
+				    &uint32_tmp, buffer);
+		safe_unpack64_array(&(*jobacct)->tres_usage_out_max,
+				    &uint32_tmp, buffer);
+		safe_unpack64_array(&(*jobacct)->tres_usage_out_max_nodeid,
+				    &uint32_tmp, buffer);
+		safe_unpack64_array(&(*jobacct)->tres_usage_out_max_taskid,
+				    &uint32_tmp, buffer);
+		safe_unpack64_array(&(*jobacct)->tres_usage_out_min,
+				    &uint32_tmp, buffer);
+		safe_unpack64_array(&(*jobacct)->tres_usage_out_min_nodeid,
+				    &uint32_tmp, buffer);
+		safe_unpack64_array(&(*jobacct)->tres_usage_out_min_taskid,
+				    &uint32_tmp, buffer);
+		safe_unpack64_array(&(*jobacct)->tres_usage_out_tot,
+				    &uint32_tmp, buffer);
+	} else if (rpc_version >= SLURM_MIN_PROTOCOL_VERSION) {
 		safe_unpack32(&uint32_tmp, buffer);
 		(*jobacct)->user_cpu_sec = uint32_tmp;
 		safe_unpack32(&uint32_tmp, buffer);
@@ -1148,10 +1206,11 @@ extern int jobacctinfo_unpack(jobacctinfo_t **jobacct,
 
 		safe_unpack32_array(&(*jobacct)->tres_ids,
 				    &(*jobacct)->tres_count, buffer);
-		slurm_unpack_list(&(*jobacct)->tres_list,
-				  slurmdb_unpack_tres_rec,
-				  slurmdb_destroy_tres_rec,
-				  buffer, rpc_version);
+		if (slurm_unpack_list(&(*jobacct)->tres_list,
+				      slurmdb_unpack_tres_rec,
+				      slurmdb_destroy_tres_rec,
+				      buffer, rpc_version) != SLURM_SUCCESS)
+			goto unpack_error;
 		safe_unpack64_array(&(*jobacct)->tres_usage_in_max,
 				    &uint32_tmp, buffer);
 		safe_unpack64_array(&(*jobacct)->tres_usage_in_max_nodeid,
@@ -1191,8 +1250,10 @@ extern int jobacctinfo_unpack(jobacctinfo_t **jobacct,
 unpack_error:
 	debug2("jobacctinfo_unpack: unpack_error: size_buf(buffer) %u",
 	       size_buf(buffer));
-	if (alloc)
+	if (alloc) {
 		jobacctinfo_destroy(*jobacct);
+		*jobacct = NULL;
+	}
 
        	return SLURM_ERROR;
 }
@@ -1209,15 +1270,15 @@ extern void jobacctinfo_aggregate(jobacctinfo_t *dest, jobacctinfo_t *from)
 
 	dest->user_cpu_sec	+= from->user_cpu_sec;
 	dest->user_cpu_usec	+= from->user_cpu_usec;
-	while (dest->user_cpu_usec >= 1E6) {
-		dest->user_cpu_sec++;
-		dest->user_cpu_usec -= 1E6;
+	if (dest->user_cpu_usec >= 1E6) {
+		dest->user_cpu_sec += dest->user_cpu_usec / 1E6;
+		dest->user_cpu_usec = dest->user_cpu_usec % (int)1E6;
 	}
 	dest->sys_cpu_sec	+= from->sys_cpu_sec;
 	dest->sys_cpu_usec	+= from->sys_cpu_usec;
-	while (dest->sys_cpu_usec >= 1E6) {
-		dest->sys_cpu_sec++;
-		dest->sys_cpu_usec -= 1E6;
+	if (dest->sys_cpu_usec >= 1E6) {
+		dest->sys_cpu_sec += dest->sys_cpu_usec / 1E6;
+		dest->sys_cpu_usec = dest->sys_cpu_usec % (int)1E6;
 	}
 	dest->act_cpufreq 	+= from->act_cpufreq;
 	if (dest->energy.consumed_energy != NO_VAL64) {
@@ -1246,3 +1307,16 @@ extern void jobacctinfo_2_stats(slurmdb_stats_t *stats, jobacctinfo_t *jobacct)
 
 	_jobacctinfo_2_stats_tres_usage(stats, jobacct);
 }
+
+extern long jobacct_gather_get_clk_tck()
+{
+	long hertz = sysconf(_SC_CLK_TCK);
+
+	if (hertz < 1) {
+		error("unable to get clock rate");
+		/* 100 is default on many systems. */
+		hertz = 100;
+	}
+
+	return hertz;
+}
