@@ -148,9 +148,9 @@ static int _set_rec(int *start, int argc, char **argv,
 
 			if (*(argv[i]+end) == '\0' &&
 			    (option == '+' || option == '-')) {
-				printf(" You didn't specify any clusters to"
-				       " %s\n",
-				       (option == '-') ? "remove" : "add");
+				fprintf(stderr,
+					" You didn't specify any clusters to %s\n",
+					(option == '-') ? "remove" : "add");
 				exit_code = 1;
 				break;
 			}
@@ -191,19 +191,20 @@ static int _set_rec(int *start, int argc, char **argv,
 				tmp_char =
 					slurmdb_federation_flags_str(
 							fed->flags);
-				printf(" Unknown federation flag used in:\n"
-				       " '%s'\n"
-				       " Valid federation flags are\n  '%s'\n",
-				       argv[i]+end, tmp_char);
+				fprintf(stderr,
+					" Unknown federation flag used in:\n '%s'\n"
+					" Valid federation flags are\n  '%s'\n",
+					argv[i]+end, tmp_char);
 				xfree(tmp_char);
 				exit_code = 1;
 			} else
 				set = 1;
 		} else {
 			exit_code = 1;
-			printf(" Unknown option: %s\n"
-			       " Use keyword 'where' to modify condition\n",
-			       argv[i]);
+			fprintf(stderr,
+				" Unknown option: %s\n"
+				" Use keyword 'where' to modify condition\n",
+				argv[i]);
 		}
 	}
 
@@ -251,7 +252,9 @@ static int _verify_federations(List name_list, bool report_existing)
 			       "Not adding.\n", name);
 			list_delete_item(itr_c);
 		} else if (!fed_rec && !report_existing) {
-			printf(" The federation %s doesn't exist.\n", name);
+			fprintf(stderr,
+				" The federation %s doesn't exist.\n",
+				name);
 			rc = SLURM_ERROR;
 		}
 	}
@@ -738,7 +741,7 @@ static int _add_clusters_to_remove(List cluster_list, const char *federation)
 		ListIterator orig_itr = list_iterator_create(cluster_list);
 
 		/* Figure out if cluster in cluster_list is already on the
-		 * federation. If if is, don't add to list to remove */
+		 * federation. If it is, don't add to list to remove */
 		while ((orig_cluster = list_next(orig_itr))) {
 			char *db_name = db_cluster->name;
 			if (*db_name == '+' || *db_name == '-')
