@@ -75,10 +75,8 @@ static int _chk_cpuinfo_str(char *buffer, char *keyword, char **valptr);
 static int _chk_cpuinfo_uint32(char *buffer, char *keyword, uint32_t *val);
 #endif
 
-static int _ranges_conv(char* lrange, char** prange, int mode);
 static int _range_to_map(char* range, uint16_t *map, uint16_t map_size,
 			 int add_threads);
-static int _map_to_range(uint16_t *map, uint16_t map_size, char** prange);
 
 bool     initialized = false;
 uint16_t procs, boards, sockets, cores, threads=1;
@@ -225,8 +223,7 @@ handle_write:
 					       HWLOC_TYPE_FILTER_KEEP_NONE);
 		hwloc_topology_set_type_filter(*topology, HWLOC_OBJ_L2CACHE,
 					       HWLOC_TYPE_FILTER_KEEP_NONE);
-		hwloc_topology_set_type_filter(*topology, HWLOC_OBJ_L3CACHE,
-					       HWLOC_TYPE_FILTER_KEEP_NONE);
+		/* need to preserve HWLOC_OBJ_L3CACHE for l3cache_as_socket */
 		hwloc_topology_set_type_filter(*topology, HWLOC_OBJ_L4CACHE,
 					       HWLOC_TYPE_FILTER_KEEP_NONE);
 		hwloc_topology_set_type_filter(*topology, HWLOC_OBJ_L5CACHE,
@@ -321,24 +318,55 @@ extern int xcpuinfo_hwloc_topo_get(
 	objtype[SOCKET] = HWLOC_OBJ_SOCKET;
 	objtype[CORE]   = HWLOC_OBJ_CORE;
 	objtype[PU]     = HWLOC_OBJ_PU;
+#if HWLOC_API_VERSION >= 0x00020000
+	if (xstrcasestr(slurm_conf.sched_params, "Ignore_NUMA")) {
+		info("SchedulerParamaters=Ignore_NUMA not supported by hwloc v2");
+	}
+#else
 	if (hwloc_get_type_depth(topology, HWLOC_OBJ_NODE) >
 	    hwloc_get_type_depth(topology, HWLOC_OBJ_SOCKET)) {
-		char *sched_params = slurm_get_sched_params();
-		if (xstrcasestr(sched_params, "Ignore_NUMA")) {
+		if (xstrcasestr(slurm_conf.sched_params, "Ignore_NUMA")) {
 			info("Ignoring NUMA nodes within a socket");
 		} else {
 			info("Considering each NUMA node as a socket");
 			objtype[SOCKET] = HWLOC_OBJ_NODE;
 		}
-		xfree(sched_params);
+	}
+#endif
+
+	if (xstrcasestr(slurm_conf.slurmd_params, "l3cache_as_socket")) {
+#if HWLOC_API_VERSION >= 0x00020000
+		objtype[SOCKET] = HWLOC_OBJ_L3CACHE;
+#else
+		error("SlurmdParameters=l3cache_as_socket requires hwloc v2");
+#endif
+	} else if (xstrcasestr(slurm_conf.slurmd_params,
+			       "numa_node_as_socket")) {
+#if HWLOC_API_VERSION >= 0x00020000
+		hwloc_obj_t numa_obj = hwloc_get_next_obj_by_type(
+			topology, HWLOC_OBJ_NODE, NULL);
+
+		if (numa_obj && numa_obj->parent) {
+			objtype[SOCKET] = numa_obj->parent->type;
+			if (get_log_level() >= LOG_LEVEL_DEBUG2) {
+				char tmp[128];
+				hwloc_obj_type_snprintf(tmp, sizeof(tmp),
+							numa_obj->parent, 0);
+				debug2("%s: numa_node_as_socket mapped to '%s'",
+				       __func__, tmp);
+			}
+		}
+#else
+		error("SlurmdParameters=numa_node_as_socket requires hwloc v2");
+#endif
 	}
 
-	/* number of objects */
-	depth = hwloc_get_type_depth(topology, HWLOC_OBJ_GROUP);
-	if (depth != HWLOC_TYPE_DEPTH_UNKNOWN) {
-		actual_boards = MAX(hwloc_get_nbobjs_by_depth(topology, depth),
-				    1);
-	}
+	/* Groups below root obj are interpreted as boards */
+	obj = hwloc_get_root_obj(topology);
+	obj = hwloc_get_next_child(topology, obj, NULL);
+	if (!hwloc_compare_types(HWLOC_OBJ_GROUP, obj->type))
+		actual_boards =
+			MAX(hwloc_get_nbobjs_by_depth(topology, obj->depth), 1);
 
 	/*
 	 * Count sockets/NUMA containing any cores.
@@ -941,36 +969,26 @@ static int _compute_block_map(uint16_t numproc,
 }
 #endif
 
-int _ranges_conv(char* lrange,char** prange,int mode);
-
-/* for testing purpose */
-/* uint16_t procs=8, sockets=2, cores=2, threads=2; */
-/* uint16_t block_map_size=8; */
-/* uint16_t block_map[] = { 0, 4, 2, 6, 1, 5, 3, 7 }; */
-/* uint16_t block_map_inv[] = { 0, 4, 2, 6, 1, 5, 3, 7 }; */
-/* xcpuinfo_abs_to_mac("0,2,4,6",&mach); */
-/* xcpuinfo_mac_to_abs(mach,&abs); */
-
 int
 xcpuinfo_init(void)
 {
 	if ( initialized )
-		return XCPUINFO_SUCCESS;
+		return SLURM_SUCCESS;
 
 	if (xcpuinfo_hwloc_topo_get(&procs,&boards,&sockets,&cores,&threads,
 				    &block_map_size,&block_map,&block_map_inv))
-		return XCPUINFO_ERROR;
+		return SLURM_ERROR;
 
 	initialized = true ;
 
-	return XCPUINFO_SUCCESS;
+	return SLURM_SUCCESS;
 }
 
 int
 xcpuinfo_fini(void)
 {
 	if ( ! initialized )
-		return XCPUINFO_SUCCESS;
+		return SLURM_SUCCESS;
 
 	initialized = false ;
 	procs = sockets = cores = threads = 0;
@@ -990,11 +1008,21 @@ xcpuinfo_fini(void)
 		xfree(hwloc_xml_whole);
 	}
 #endif
-	return XCPUINFO_SUCCESS;
+	return SLURM_SUCCESS;
 }
 
-int
-xcpuinfo_abs_to_mac(char* lrange,char** prange)
+/*
+ * Convert an abstract core range string into a machine-specific CPU range
+ * string. Abstract id to machine id conversion is done using block_map.
+ * When a core is set in the input, all its threads will be set in the output.
+ *
+ * Inverse of xcpuinfo_mac_to_abs.
+ *
+ * Input:  lrange - abstract/logical core spec string.
+ * Output: prange - machine/local/physical CPU spec string.
+ *         return code - SLURM_SUCCESS if no error, otherwise SLURM_ERROR.
+ */
+int xcpuinfo_abs_to_mac(char *lrange, char **prange)
 {
 	static int total_cores = -1, total_cpus = -1;
 	bitstr_t* absmap = NULL;
@@ -1026,11 +1054,11 @@ xcpuinfo_abs_to_mac(char* lrange,char** prange)
 	/* mapping abstract id to machine id using conf->block_map */
 	for (icore = 0; icore < total_cores; icore++) {
 		if (bit_test(absmap, icore)) {
-			for (ithread = 0; ithread<conf->threads; ithread++) {
-				absid  = icore*conf->threads + ithread;
+			for (ithread = 0; ithread < conf->threads; ithread++) {
+				absid = icore * conf->threads + ithread;
 				absid %= total_cpus;
 
-				macid  = conf->block_map[absid];
+				macid = conf->block_map[absid];
 				macid %= total_cpus;
 
 				bit_set(macmap, macid);
@@ -1048,15 +1076,97 @@ end_it:
 	FREE_NULL_BITMAP(macmap);
 
 	if (rc != SLURM_SUCCESS)
-		info("_abs_to_mac failed");
+		error("%s: failed", __func__);
 
 	return rc;
 }
 
-int
-xcpuinfo_mac_to_abs(char* lrange,char** prange)
+/*
+ * Convert a machine-specific CPU range string into an abstract core range
+ * string. Machine id to abstract id conversion is done using block_map_inv.
+ * When a single thread in a core is set in the input, the corresponding core
+ * will be set in its output.
+ *
+ * Inverse of xcpuinfo_abs_to_mac.
+ *
+ * Input:  in_range - machine/local/physical CPU range string.
+ * Output: out_range - abstract/logical core range string. Caller should xfree()
+ *         return code - SLURM_SUCCESS if no error, otherwise SLURM_ERROR.
+ */
+int xcpuinfo_mac_to_abs(char *in_range, char **out_range)
 {
-	return _ranges_conv(lrange,prange,1);
+	static int total_cores = -1, total_cpus = -1;
+	bitstr_t *macmap = NULL;
+	bitstr_t *absmap = NULL;
+	bitstr_t *absmap_core = NULL;
+	int rc = SLURM_SUCCESS;
+
+	if (total_cores == -1) {
+		total_cores = conf->sockets * conf->cores;
+		total_cpus  = conf->block_map_size;
+	}
+
+	/* allocate bitmaps */
+	macmap = bit_alloc(total_cpus);
+	absmap = bit_alloc(total_cpus);
+	absmap_core = bit_alloc(total_cores);
+
+	if (!macmap || !absmap) {
+		rc = SLURM_ERROR;
+		goto end_it;
+	}
+
+	/* string to bitmap conversion */
+	if (bit_unfmt(macmap, in_range)) {
+		rc = SLURM_ERROR;
+		goto end_it;
+	}
+
+	/* mapping machine id to abstract id using conf->block_map_inv */
+	for (int icore = 0; icore < total_cores; icore++) {
+		for (int ithread = 0; ithread < conf->threads; ithread++) {
+			int absid, macid;
+			macid = (icore * conf->threads) + ithread;
+			macid %= total_cpus;
+
+			/* Skip this machine CPU id if not in in_range */
+			if (!bit_test(macmap, macid))
+				continue;
+
+			absid = conf->block_map_inv[macid];
+			absid %= total_cpus;
+
+			bit_set(absmap, absid);
+		}
+	}
+
+	/* condense abstract CPU bitmap into an abstract core bitmap */
+	for (int icore = 0; icore < total_cores; icore++) {
+		for (int ithread = 0; ithread < conf->threads; ithread++) {
+			int icpu = (icore * conf->threads) + ithread;
+			icpu %= total_cpus;
+
+			if (bit_test(absmap, icpu)) {
+				bit_set(absmap_core, icore);
+				break;
+			}
+		}
+	}
+
+	/* convert abstract core bitmap to range string */
+	*out_range = (char*)xmalloc(total_cores * 6);
+	bit_fmt(*out_range, total_cores * 6, absmap_core);
+
+	/* free unused bitmaps */
+end_it:
+	FREE_NULL_BITMAP(macmap);
+	FREE_NULL_BITMAP(absmap);
+	FREE_NULL_BITMAP(absmap_core);
+
+	if (rc != SLURM_SUCCESS)
+		error("%s failed", __func__);
+
+	return rc;
 }
 
 int
@@ -1068,57 +1178,6 @@ xcpuinfo_abs_to_map(char* lrange,uint16_t **map,uint16_t *map_size)
 	return _range_to_map(lrange,*map,*map_size,1);
 }
 
-int
-xcpuinfo_map_to_mac(uint16_t *map,uint16_t map_size,char** range)
-{
-	return _map_to_range(map,map_size,range);
-}
-
-int
-xcpuinfo_mac_to_map(char* lrange,uint16_t **map,uint16_t *map_size)
-{
-	*map_size = block_map_size;
-	*map = (uint16_t*) xmalloc(block_map_size*sizeof(uint16_t));
-	/* machine range already includes the hyperthreads */
-	return _range_to_map(lrange,*map,*map_size,0);
-}
-
-int
-xcpuinfo_absmap_to_macmap(uint16_t *amap,uint16_t amap_size,
-			  uint16_t **bmap,uint16_t *bmap_size)
-{
-	/* int i; */
-
-	/* abstract to machine conversion using block map */
-	uint16_t *map_out;
-
-	*bmap_size = amap_size;
-	map_out = (uint16_t*) xmalloc(amap_size*sizeof(uint16_t));
-	*bmap = map_out;
-
-	return XCPUINFO_SUCCESS;
-}
-
-int
-xcpuinfo_macmap_to_absmap(uint16_t *amap,uint16_t amap_size,
-			  uint16_t **bmap,uint16_t *bmap_size)
-{
-	int i;
-
-	/* machine to abstract conversion using inverted block map */
-	uint16_t *cmap;
-	cmap = block_map_inv;
-	*bmap_size = amap_size;
-	*bmap = (uint16_t*) xmalloc(amap_size*sizeof(uint16_t));
-	for( i = 0 ; i < amap_size ; i++) {
-		if ( amap[i] )
-			(*bmap)[cmap[i]]=1;
-		else
-			(*bmap)[cmap[i]]=0;
-	}
-	return XCPUINFO_SUCCESS;
-}
-
 /*
  * set to 1 each element of already allocated map of size
  * map_size if they are present in the input range
@@ -1193,123 +1252,8 @@ _range_to_map(char* range,uint16_t *map,uint16_t map_size,int add_threads)
 
 	if ( bad_nb > 0 ) {
 		/* bad format for input range */
-		return XCPUINFO_ERROR;
+		return SLURM_ERROR;
 	}
 
-	return XCPUINFO_SUCCESS;
-}
-
-
-/*
- * allocate and build a range of ids using an input map
- * having printable element set to 1
- */
-static int
-_map_to_range(uint16_t *map,uint16_t map_size,char** prange)
-{
-	size_t len;
-	int num_fl=0;
-	int con_fl=0;
-
-	char *str;
-
-	uint16_t start=0,end=0,i;
-
-	str = xstrdup("");
-	for ( i = 0 ; i < map_size ; i++ ) {
-
-		if ( map[i] ) {
-			num_fl=1;
-			end=i;
-			if ( !con_fl ) {
-				start=end;
-				con_fl=1;
-			}
-		}
-		else if ( num_fl ) {
-			if ( start < end ) {
-				xstrfmtcat(str, "%u-%u,", start, end);
-			}
-			else {
-				xstrfmtcat(str, "%u,", start);
-			}
-			con_fl = num_fl = 0;
-		}
-	}
-	if ( num_fl ) {
-		if ( start < end ) {
-			xstrfmtcat(str, "%u-%u,", start, end);
-		}
-		else {
-			xstrfmtcat(str, "%u,", start);
-		}
-	}
-
-	len = strlen(str);
-	if ( len > 0 ) {
-		str[len-1]='\0';
-	}
-	else {
-		xfree(str);
-		return XCPUINFO_ERROR;
-	}
-
-	if ( prange != NULL )
-		*prange = str;
-	else
-		xfree(str);
-
-	return XCPUINFO_SUCCESS;
-}
-
-/*
- * convert a range into an other one according to
- * a modus operandi being 0 or 1 for abstract to machine
- * or machine to abstract representation of cores
- */
-static int
-_ranges_conv(char* lrange,char** prange,int mode)
-{
-	int fstatus;
-	int i;
-	uint16_t *amap;
-	uint16_t *map;
-	uint16_t *map_out;
-
-	/* init internal data if not already done */
-	if ( xcpuinfo_init() != XCPUINFO_SUCCESS )
-		return XCPUINFO_ERROR;
-
-	if ( mode ) {
-		/* machine to abstract conversion */
-		amap = block_map_inv;
-	}
-	else {
-		/* abstract to machine conversion */
-		amap = block_map;
-	}
-
-	/* allocate map for local work */
-	map = (uint16_t*) xmalloc(block_map_size*sizeof(uint16_t));
-	map_out = (uint16_t*) xmalloc(block_map_size*sizeof(uint16_t));
-
-	/* extract the input map */
-	fstatus = _range_to_map(lrange,map,block_map_size,!mode);
-	if ( fstatus ) {
-		goto exit;
-	}
-
-	/* do the conversion (see src/slurmd/slurmd/get_mach_stat.c) */
-	for( i = 0 ; i < block_map_size ; i++) {
-		if ( map[i] )
-			map_out[amap[i]]=1;
-	}
-
-	/* build the ouput range */
-	fstatus = _map_to_range(map_out,block_map_size,prange);
-
-exit:
-	xfree(map);
-	xfree(map_out);
-	return fstatus;
+	return SLURM_SUCCESS;
 }
