@@ -70,6 +70,10 @@
 #  endif
 #endif
 
+#ifdef WITH_SELINUX
+#include <selinux/selinux.h>
+#endif
+
 #include "slurm/slurm_errno.h"
 
 #include "src/common/cbuf.h"
@@ -81,8 +85,9 @@
 #include "src/common/hostlist.h"
 #include "src/common/log.h"
 #include "src/common/macros.h"
-#include "src/common/node_select.h"
 #include "src/common/plugstack.h"
+#include "src/common/reverse_tree.h"
+#include "src/common/select.h"
 #include "src/common/slurm_acct_gather_profile.h"
 #include "src/common/slurm_cred.h"
 #include "src/common/slurm_jobacct_gather.h"
@@ -94,16 +99,14 @@
 #include "src/common/xmalloc.h"
 #include "src/common/xsignal.h"
 #include "src/common/xstring.h"
+#include "src/common/x11_util.h"
 
 #include "src/slurmd/slurmd/slurmd.h"
 
 #include "src/slurmd/common/core_spec_plugin.h"
 #include "src/slurmd/common/fname.h"
 #include "src/slurmd/common/job_container_plugin.h"
-#include "src/slurmd/common/log_ctld.h"
 #include "src/slurmd/common/proctrack.h"
-#include "src/slurmd/common/run_script.h"
-#include "src/slurmd/common/reverse_tree.h"
 #include "src/slurmd/common/set_oomadj.h"
 #include "src/slurmd/common/slurmd_cgroup.h"
 #include "src/slurmd/common/task_plugin.h"
@@ -123,14 +126,6 @@
 #define RETRY_DELAY 15		/* retry every 15 seconds */
 #define MAX_RETRY   240		/* retry 240 times (one hour max) */
 
-struct priv_state {
-	uid_t	saved_uid;
-	gid_t	saved_gid;
-	gid_t *	gid_list;
-	int	ngids;
-	char	saved_cwd [4096];
-};
-
 step_complete_t step_complete = {
 	PTHREAD_COND_INITIALIZER,
 	PTHREAD_MUTEX_INITIALIZER,
@@ -164,14 +159,11 @@ static int sig_array[] = {SIGTERM, 0};
 static bool _access(const char *path, int modes, uid_t uid,
 		    int ngids, gid_t *gids);
 static void _send_launch_failure(launch_tasks_request_msg_t *,
-				 slurm_addr_t *, uid_t, int, uint16_t);
+				 slurm_addr_t *, int, uint16_t);
 static int  _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized);
 static int  _become_user(stepd_step_rec_t *job, struct priv_state *ps);
 static void  _set_prio_process (stepd_step_rec_t *job);
 static int  _setup_normal_io(stepd_step_rec_t *job);
-static int  _drop_privileges(stepd_step_rec_t *job, bool do_setuid,
-			     struct priv_state *state, bool get_list);
-static int  _reclaim_privileges(struct priv_state *state);
 static void _send_launch_resp(stepd_step_rec_t *job, int rc);
 static int  _slurmd_job_log_init(stepd_step_rec_t *job);
 static void _wait_for_io(stepd_step_rec_t *job);
@@ -192,7 +184,8 @@ static void *_x11_signal_handler(void *arg);
  * Batch job management prototypes:
  */
 static char * _make_batch_dir(stepd_step_rec_t *job);
-static char * _make_batch_script(batch_job_launch_msg_t *msg, char *path);
+static int _make_batch_script(batch_job_launch_msg_t *msg,
+			      stepd_step_rec_t *job);
 static int    _send_complete_batch_script_msg(stepd_step_rec_t *job,
 					      int err, int status);
 
@@ -201,8 +194,7 @@ static int    _send_complete_batch_script_msg(stepd_step_rec_t *job,
  */
 extern stepd_step_rec_t *
 mgr_launch_tasks_setup(launch_tasks_request_msg_t *msg, slurm_addr_t *cli,
-		       uid_t cli_uid, slurm_addr_t *self,
-		       uint16_t protocol_version)
+		       slurm_addr_t *self, uint16_t protocol_version)
 {
 	stepd_step_rec_t *job = NULL;
 
@@ -213,8 +205,7 @@ mgr_launch_tasks_setup(launch_tasks_request_msg_t *msg, slurm_addr_t *cli,
 		 * reset in _send_launch_failure.
 		 */
 		int fail = errno;
-		_send_launch_failure(msg, cli, cli_uid, errno,
-				     protocol_version);
+		_send_launch_failure(msg, cli, errno, protocol_version);
 		errno = fail;
 		return NULL;
 	}
@@ -255,8 +246,9 @@ _send_srun_resp_msg(slurm_msg_t *resp_msg, uint32_t nnodes)
 		if (!max_retry)
 			max_retry = (nnodes / 1024) + 5;
 
-		debug("%s: %d/%d failed to send msg type %u: %m",
-		      __func__, retry, max_retry, resp_msg->msg_type);
+		debug("%s: %d/%d failed to send msg type %s: %m",
+		      __func__, retry, max_retry,
+		      rpc_num2string(resp_msg->msg_type));
 
 		if (retry >= max_retry)
 			break;
@@ -348,16 +340,23 @@ static uint32_t _get_exit_code(stepd_step_rec_t *job)
 	return step_rc;
 }
 
+static char *_batch_script_path(stepd_step_rec_t *job)
+{
+	return xstrdup_printf("%s/%s", job->batchdir, "slurm_script");
+}
+
 /*
  * Send batch exit code to slurmctld. Non-zero rc will DRAIN the node.
  */
 extern void
 batch_finish(stepd_step_rec_t *job, int rc)
 {
+	char *script = _batch_script_path(job);
 	step_complete.step_rc = _get_exit_code(job);
 
-	if (job->argv[0] && (unlink(job->argv[0]) < 0))
-		error("unlink(%s): %m", job->argv[0]);
+	if (unlink(script) < 0)
+		error("unlink(%s): %m", script);
+	xfree(script);
 
 	if (job->aborted) {
 		if (job->step_id.step_id != SLURM_BATCH_SCRIPT)
@@ -390,16 +389,10 @@ stepd_step_rec_t *
 mgr_launch_batch_job_setup(batch_job_launch_msg_t *msg, slurm_addr_t *cli)
 {
 	stepd_step_rec_t *job = NULL;
-	char *err_msg = NULL;
 
 	if (!(job = batch_stepd_step_rec_create(msg))) {
-		xstrfmtcat(err_msg,
-			   "batch_stepd_step_rec_create() failed for job %u on %s: %s",
-			   msg->job_id, conf->hostname,
-			   slurm_strerror(errno));
-		(void) log_ctld(LOG_LEVEL_ERROR, err_msg);
-		error("%s", err_msg);
-		xfree(err_msg);
+		error("batch_stepd_step_rec_create() failed for job %u on %s: %s",
+		      msg->job_id, conf->hostname, slurm_strerror(errno));
 		return NULL;
 	}
 
@@ -409,9 +402,8 @@ mgr_launch_batch_job_setup(batch_job_launch_msg_t *msg, slurm_addr_t *cli)
 
 	xfree(job->argv[0]);
 
-	if ((job->argv[0] = _make_batch_script(msg, job->batchdir)) == NULL) {
+	if (_make_batch_script(msg, job))
 		goto cleanup;
-	}
 
 	/* this is the new way of setting environment variables */
 	env_array_for_batch_job(&job->env, msg, conf->node_name);
@@ -424,13 +416,8 @@ mgr_launch_batch_job_setup(batch_job_launch_msg_t *msg, slurm_addr_t *cli)
 	return job;
 
 cleanup:
-	xstrfmtcat(err_msg,
-		   "batch script setup failed for job %u on %s: %s",
-		   msg->job_id, conf->hostname,
-		   slurm_strerror(errno));
-	(void) log_ctld(LOG_LEVEL_ERROR, err_msg);
-	error("%s", err_msg);
-	xfree(err_msg);
+	error("batch script setup failed for job %u on %s: %s",
+	      msg->job_id, conf->hostname, slurm_strerror(errno));
 
 	if (job->aborted)
 		verbose("job %u abort complete", job->step_id.job_id);
@@ -460,7 +447,7 @@ _setup_normal_io(stepd_step_rec_t *job)
 	 * descriptors (which may be connected to files), then
 	 * reclaim privileges.
 	 */
-	if (_drop_privileges(job, true, &sprivs, true) < 0)
+	if (drop_privileges(job, true, &sprivs, true) < 0)
 		return ESLURMD_SET_UID_OR_GID_ERROR;
 
 	if (io_init_tasks_stdio(job) != SLURM_SUCCESS) {
@@ -574,7 +561,7 @@ _setup_normal_io(stepd_step_rec_t *job)
 	}
 
 claim:
-	if (_reclaim_privileges(&sprivs) < 0) {
+	if (reclaim_privileges(&sprivs) < 0) {
 		error("sete{u/g}id(%lu/%lu): %m",
 		      (u_long) sprivs.saved_uid, (u_long) sprivs.saved_gid);
 	}
@@ -586,19 +573,6 @@ claim:
 	return rc;
 }
 
-static int
-_setup_user_managed_io(stepd_step_rec_t *job)
-{
-	srun_info_t *srun;
-
-	if ((srun = list_peek(job->sruns)) == NULL) {
-		error("_setup_user_managed_io: no clients!");
-		return SLURM_ERROR;
-	}
-
-	return user_managed_io_client_connect(job->node_tasks, srun, job->task);
-}
-
 static void
 _random_sleep(stepd_step_rec_t *job)
 {
@@ -923,12 +897,43 @@ static void _set_job_state(stepd_step_rec_t *job, slurmstepd_state_t new_state)
 	slurm_mutex_unlock(&job->state_mutex);
 }
 
+static bool _need_join_container()
+{
+	/*
+	 * To avoid potential problems with the job_container/tmpfs and
+	 * home_xauthority, don't join the container to create the xauthority
+	 * file when it is set.
+	 */
+	if ((xstrcasestr(slurm_conf.job_container_plugin, "tmpfs")) &&
+	    (!xstrcasestr(slurm_conf.x11_params, "home_xauthority"))) {
+		return true;
+	}
+
+	return false;
+}
+
+static void _shutdown_x11_forward(stepd_step_rec_t *job)
+{
+	struct priv_state sprivs = { 0 };
+
+	if (drop_privileges(job, true, &sprivs, false) < 0) {
+		error("%s: Unable to drop privileges", __func__);
+		return;
+	}
+
+	if (shutdown_x11_forward(job) != SLURM_SUCCESS)
+		error("%s: x11 forward shutdown failed", __func__);
+
+	if (reclaim_privileges(&sprivs) < 0)
+		error("%s: Unable to reclaim privileges", __func__);
+}
+
 static void *_x11_signal_handler(void *arg)
 {
 	stepd_step_rec_t *job = (stepd_step_rec_t *) arg;
-	struct priv_state sprivs = { 0 };
-	int sig;
+	int sig, status;
 	sigset_t set;
+	pid_t cpid, pid;
 
 	(void) pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
 	(void) pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
@@ -941,21 +946,64 @@ static void *_x11_signal_handler(void *arg)
 		switch (sig) {
 		case SIGTERM:	/* kill -15 */
 			debug("Terminate signal (SIGTERM) received");
-			if (_drop_privileges(job, true, &sprivs, false) < 0) {
-				error("Unable to drop privileges");
+			if (!_need_join_container()) {
+				_shutdown_x11_forward(job);
 				return NULL;
 			}
-			shutdown_x11_forward(job);
-			if (_reclaim_privileges(&sprivs) < 0)
-				error("Unable to reclaim privileges");
-			return NULL;	/* Normal termination */
-			break;
+			if ((cpid = fork()) == 0) {
+				if (container_g_join(job->step_id.job_id,
+						     job->uid) !=
+				    SLURM_SUCCESS) {
+					error("%s: cannot join container",
+					      __func__);
+					_exit(1);
+				}
+				_shutdown_x11_forward(job);
+				_exit(0);
+			} else if (cpid < 0) {
+				error("%s: fork: %m", __func__);
+			} else {
+				pid = waitpid(cpid, &status, 0);
+				if (pid < 0)
+					error("%s: waitpid failed: %m",
+					      __func__);
+				else if (!WIFEXITED(status))
+					error("%s: child terminated abnormally",
+					      __func__);
+				else if (WEXITSTATUS(status))
+					error("%s: child returned non-zero",
+					      __func__);
+			}
+			return NULL;
 		default:
 			error("Invalid signal (%d) received", sig);
 		}
 	}
 }
 
+static int _set_xauthority(stepd_step_rec_t *job)
+{
+	struct priv_state sprivs = { 0 };
+
+	if (drop_privileges(job, true, &sprivs, false) < 0) {
+		error("%s: Unable to drop privileges before xauth", __func__);
+		return SLURM_ERROR;
+	}
+
+	if (x11_set_xauth(job->x11_xauthority, job->x11_magic_cookie,
+			  job->x11_display)) {
+		error("%s: failed to run xauth", __func__);
+		return SLURM_ERROR;
+	}
+
+	if (reclaim_privileges(&sprivs) < 0) {
+		error("%s: Unable to reclaim privileges after xauth", __func__);
+		return SLURM_ERROR;
+	}
+
+	return SLURM_SUCCESS;
+}
+
 static int _spawn_job_container(stepd_step_rec_t *job)
 {
 	jobacctinfo_t *jobacct = NULL;
@@ -966,6 +1014,19 @@ static int _spawn_job_container(stepd_step_rec_t *job)
 	int rc = SLURM_SUCCESS;
 	uint32_t jobid;
 
+#ifdef HAVE_NATIVE_CRAY
+	if (job->het_job_id && (job->het_job_id != NO_VAL))
+		jobid = job->het_job_id;
+	else
+		jobid = job->step_id.job_id;
+#else
+	jobid = job->step_id.job_id;
+#endif
+	if (container_g_stepd_create(jobid, job->uid)) {
+		error("%s: container_g_stepd_create(%u): %m", __func__, jobid);
+		return SLURM_ERROR;
+	}
+
 	debug2("%s: Before call to spank_init()", __func__);
 	if (spank_init(job) < 0) {
 		error("%s: Plugin stack initialization failed.", __func__);
@@ -975,10 +1036,10 @@ static int _spawn_job_container(stepd_step_rec_t *job)
 	}
 	debug2("%s: After call to spank_init()", __func__);
 
-	set_oom_adj(0);	/* the tasks may be killed by OOM */
 	if (task_g_pre_setuid(job)) {
 		error("%s: Failed to invoke task plugins: one of "
 		      "task_p_pre_setuid functions returned error", __func__);
+		return SLURM_ERROR;
 	}
 
 	acct_gather_profile_g_task_start(0);
@@ -986,7 +1047,7 @@ static int _spawn_job_container(stepd_step_rec_t *job)
 	if (job->x11) {
 		struct priv_state sprivs = { 0 };
 
-		if (_drop_privileges(job, true, &sprivs, false) < 0) {
+		if (drop_privileges(job, true, &sprivs, false) < 0) {
 			error ("Unable to drop privileges");
 			return SLURM_ERROR;
 		}
@@ -995,7 +1056,7 @@ static int _spawn_job_container(stepd_step_rec_t *job)
 			error("x11 port forwarding setup failed");
 			_exit(127);
 		}
-		if (_reclaim_privileges(&sprivs) < 0) {
+		if (reclaim_privileges(&sprivs) < 0) {
 			error ("Unable to reclaim privileges");
 			return SLURM_ERROR;
 		}
@@ -1009,12 +1070,54 @@ static int _spawn_job_container(stepd_step_rec_t *job)
 		      job->x11_xauthority);
 	}
 
+	/*
+	 * When using job_container/tmpfs we need to get into
+	 * the correct namespace or .Xauthority won't be visible
+	 * in /tmp from inside the job.
+	 */
+	if (job->x11) {
+		if (_need_join_container()) {
+			/*
+			 * The fork is necessary because we cannot join a
+			 * namespace if we are multithreaded. Also we need to
+			 * wait for the child to end before proceeding or there
+			 * can be a timing race with srun starting X11 apps very
+			 * fast.
+			 */
+			pid = fork();
+			if (pid == 0) {
+				if (container_g_join(jobid, job->uid) !=
+				    SLURM_SUCCESS)
+					_exit(1);
+				_exit(_set_xauthority(job));
+			} else if (pid < 0) {
+				error("fork: %m");
+				rc = SLURM_ERROR;
+			}
+			if ((waitpid(pid, &status, 0) != pid) ||
+			    WEXITSTATUS(status)) {
+				error("%s: Xauthority setup failed", __func__);
+				rc = SLURM_ERROR;
+			}
+		} else {
+			rc = _set_xauthority(job);
+		}
+
+		if (rc != SLURM_SUCCESS) {
+			_set_job_state(job, SLURMSTEPD_STEP_ENDING);
+			close_slurmd_conn();
+			goto fail1;
+		}
+	}
+
 	pid = fork();
 	if (pid == 0) {
 		setpgid(0, 0);
 		setsid();
+		set_oom_adj(0);	/* the tasks may be killed by OOM */
 		acct_gather_profile_g_child_forked();
 		_unblock_signals();
+
 		/*
 		 * Need to exec() something for proctrack/linuxproc to
 		 * work, it will not keep a process named "slurmstepd"
@@ -1049,14 +1152,6 @@ static int _spawn_job_container(stepd_step_rec_t *job)
 	jobacct_id.job    = job;
 	jobacct_gather_set_proctrack_container_id(job->cont_id);
 	jobacct_gather_add_task(pid, &jobacct_id, 1);
-#ifdef HAVE_NATIVE_CRAY
-	if (job->het_job_id && (job->het_job_id != NO_VAL))
-		jobid = job->het_job_id;
-	else
-		jobid = job->step_id.job_id;
-#else
-	jobid = job->step_id.job_id;
-#endif
 	container_g_add_cont(jobid, job->cont_id);
 
 	_set_job_state(job, SLURMSTEPD_STEP_RUNNING);
@@ -1124,8 +1219,10 @@ static int _spawn_job_container(stepd_step_rec_t *job)
 	task_g_post_step(job);
 
 fail1:
-	if (x11_signal_handler_thread)
+	if (x11_signal_handler_thread) {
 		(void) pthread_kill(x11_signal_handler_thread, SIGTERM);
+		pthread_join(x11_signal_handler_thread, NULL);
+	}
 
 	debug2("%s: Before call to spank_fini()", __func__);
 	if (spank_fini(job) < 0)
@@ -1154,7 +1251,7 @@ job_manager(stepd_step_rec_t *job)
 {
 	int  rc = SLURM_SUCCESS;
 	bool io_initialized = false;
-	char *err_msg = NULL;
+	char *oom_val_str;
 
 	debug3("Entered job_manager for %ps pid=%d",
 	       &job->step_id, job->jmgr_pid);
@@ -1164,6 +1261,15 @@ job_manager(stepd_step_rec_t *job)
 		debug ("Unable to set dumpable to 1");
 #endif /* PR_SET_DUMPABLE */
 
+	/*
+	 * Set oom_score_adj of this slurmstepd to -1000 to avoid OOM killing
+	 * us. If we were killed at this pont due to other steps OOMing, no
+	 * cleanup would happen, leaving for example cgroup stray directories if
+	 * cgroup plugins were initialized.
+	 */
+	set_oom_adj(-1000);
+	debug("Setting slurmstepd(%d) oom_score_adj to -1000", getpid());
+
 	/*
 	 * Run acct_gather_conf_init() now so we don't drop permissions on any
 	 * of the gather plugins.
@@ -1179,20 +1285,37 @@ job_manager(stepd_step_rec_t *job)
 	    (acct_gather_profile_init() != SLURM_SUCCESS)	||
 	    (slurm_cred_init() != SLURM_SUCCESS)		||
 	    (job_container_init() != SLURM_SUCCESS)		||
-	    (gres_plugin_init() != SLURM_SUCCESS)) {
+	    (gres_init() != SLURM_SUCCESS)) {
 		rc = SLURM_PLUGIN_NAME_INVALID;
 		goto fail1;
 	}
+
+	/*
+	 * Readjust this slurmstepd oom_score_adj now that we've loaded the
+	 * task plugin. If the environment variable SLURMSTEPD_OOM_ADJ is set
+	 * and is a valid number (from -1000 to 1000) set the score to that
+	 * value. Note that if the value is -1000 we will do nothing as that was
+	 * already done before.
+	 */
+	if ((oom_val_str = getenv("SLURMSTEPD_OOM_ADJ"))) {
+		int oom_val = atoi(oom_val_str);
+		if ((oom_val > -1000) && (oom_val <= 1000)) {
+			debug("Setting slurmstepd oom_score_adj from env to %d",
+			      oom_val);
+			set_oom_adj(oom_val);
+		}
+	}
+
 	if (!job->batch && (job->step_id.step_id != SLURM_EXTERN_CONT) &&
 	    (job->step_id.step_id != SLURM_INTERACTIVE_STEP) &&
-	    (mpi_hook_slurmstepd_init(&job->env) != SLURM_SUCCESS)) {
+	    (mpi_process_env(&job->env) != SLURM_SUCCESS)) {
 		rc = SLURM_MPI_PLUGIN_NAME_INVALID;
 		goto fail1;
 	}
 
 	if (!job->batch && (job->step_id.step_id != SLURM_EXTERN_CONT) &&
 	    (job->step_id.step_id != SLURM_INTERACTIVE_STEP) &&
-	    (switch_g_job_preinit(job->switch_job) < 0)) {
+	    (switch_g_job_preinit(job) < 0)) {
 		rc = ESLURM_INTERCONNECT_FAILURE;
 		goto fail1;
 	}
@@ -1225,14 +1348,9 @@ job_manager(stepd_step_rec_t *job)
 
 	/* fork necessary threads for MPI */
 	if (!job->batch && (job->step_id.step_id != SLURM_INTERACTIVE_STEP) &&
-	    (mpi_hook_slurmstepd_prefork(job, &job->env) != SLURM_SUCCESS)) {
-		error("Failed mpi_hook_slurmstepd_prefork");
+	    (mpi_g_slurmstepd_prefork(job, &job->env) != SLURM_SUCCESS)) {
+		error("Failed mpi_g_slurmstepd_prefork");
 		rc = SLURM_ERROR;
-		xstrfmtcat(err_msg,
-			   "mpi_hook_slurmstepd_prefork failure for %ps on %s",
-			   &job->step_id, conf->hostname);
-		(void) log_ctld(LOG_LEVEL_ERROR, err_msg);
-		xfree(err_msg);
 		goto fail3;
 	}
 
@@ -1245,14 +1363,11 @@ job_manager(stepd_step_rec_t *job)
 	if (!job->batch && (job->step_id.step_id != SLURM_INTERACTIVE_STEP) &&
 	    (job->node_tasks > 1) &&
 	    (job->accel_bind_type || job->tres_bind)) {
-		uint64_t gpu_cnt, mic_cnt, nic_cnt;
-		gpu_cnt = gres_plugin_step_count(job->step_gres_list, "gpu");
-		mic_cnt = gres_plugin_step_count(job->step_gres_list, "mic");
-		nic_cnt = gres_plugin_step_count(job->step_gres_list, "nic");
+		uint64_t gpu_cnt, nic_cnt;
+		gpu_cnt = gres_step_count(job->step_gres_list, "gpu");
+		nic_cnt = gres_step_count(job->step_gres_list, "nic");
 		if ((gpu_cnt <= 1) || (gpu_cnt == NO_VAL64))
 			job->accel_bind_type &= (~ACCEL_BIND_CLOSEST_GPU);
-		if ((mic_cnt <= 1) || (mic_cnt == NO_VAL64))
-			job->accel_bind_type &= (~ACCEL_BIND_CLOSEST_MIC);
 		if ((nic_cnt <= 1) || (nic_cnt == NO_VAL64))
 			job->accel_bind_type &= (~ACCEL_BIND_CLOSEST_NIC);
 		if (job->accel_bind_type == ACCEL_BIND_VERBOSE)
@@ -1351,8 +1466,7 @@ fail2:
 	/*
 	 * Wait for io thread to complete (if there is one)
 	 */
-	if (!job->batch && io_initialized &&
-	    ((job->flags & LAUNCH_USER_MANAGED_IO) == 0))
+	if (!job->batch && io_initialized)
 		_wait_for_io(job);
 
 	/*
@@ -1374,7 +1488,7 @@ fail2:
 	if (!job->batch && (job->step_id.step_id != SLURM_INTERACTIVE_STEP) &&
 	    job->tres_freq) {
 		if (getuid() == (uid_t) 0)
-			gres_plugin_step_hardware_fini();
+			gres_g_step_hardware_fini();
 		else
 			error("%s: invalid permissions: cannot uninitialize GRES hardware unless Slurmd was started as root",
 			      __func__);
@@ -1407,7 +1521,8 @@ fail1:
 	 */
 	_set_job_state(job, SLURMSTEPD_STEP_ENDING);
 	if (rc != 0) {
-		error("job_manager exiting abnormally, rc = %d", rc);
+		error("%s: exiting abnormally: %s",
+		      __func__, slurm_strerror(rc));
 		_send_launch_resp(job, rc);
 	}
 
@@ -1432,36 +1547,47 @@ static int _pre_task_child_privileged(
 	int setwd = 0; /* set working dir */
 	int rc = 0;
 
-	if (_reclaim_privileges(sp) < 0)
+	if (reclaim_privileges(sp) < 0)
 		return SLURM_ERROR;
 
+	set_oom_adj(0); /* the tasks may be killed by OOM */
+
 #ifndef HAVE_NATIVE_CRAY
-	/* Add job's pid to job container */
+	if (!(job->flags & LAUNCH_NO_ALLOC)) {
+		/* Add job's pid to job container, if a normal job */
+		if (container_g_join(job->step_id.job_id, job->uid)) {
+			error("container_g_join failed: %u",
+			      job->step_id.job_id);
+			exit(1);
+		}
 
-	if (container_g_join(job->step_id.job_id, job->uid)) {
-		error("container_g_join failed: %u", job->step_id.job_id);
-		exit(1);
+		/*
+		 * tmpfs job container plugin changes the working directory
+		 * back to root working directory, so change it back to users
+		 * but after dropping privillege
+		 */
+		setwd = 1;
 	}
-
-	/*
-	 * tmpfs job container plugin changes the working directory
-	 * back to root working directory, so change it back to users
-	 * but after dropping privillege
-	 */
-	setwd = 1;
 #endif
 
 	if (spank_task_privileged(job, taskid) < 0)
 		return error("spank_task_init_privileged failed");
 
 	/* sp->gid_list should already be initialized */
-	rc = _drop_privileges(job, true, sp, false);
+	rc = drop_privileges(job, true, sp, false);
 	if (rc) {
-		error ("_drop_privileges: %m");
+		error ("drop_privileges: %m");
 		return rc;
 	}
 
-	if (setwd) {
+	if (job->container) {
+		/* Container jobs must start in the correct directory */
+		if (chdir(job->cwd) < 0) {
+			error("couldn't chdir to `%s': %m", job->cwd);
+			return errno;
+		}
+		debug2("%s: chdir(%s) success", __func__, job->cwd);
+	} else if (setwd) {
 		if (chdir(job->cwd) < 0) {
 			error("couldn't chdir to `%s': %m: going to /tmp instead",
 			      job->cwd);
@@ -1470,7 +1596,6 @@ static int _pre_task_child_privileged(
 				return SLURM_ERROR;
 			}
 		}
-
 	}
 
 	return rc;
@@ -1489,14 +1614,11 @@ static struct exec_wait_info * _exec_wait_info_create (int i)
 	int fdpair[2];
 	struct exec_wait_info * e;
 
-	if (pipe (fdpair) < 0) {
+	if (pipe2(fdpair, O_CLOEXEC) < 0) {
 		error ("_exec_wait_info_create: pipe: %m");
 		return NULL;
 	}
 
-	fd_set_close_on_exec(fdpair[0]);
-	fd_set_close_on_exec(fdpair[1]);
-
 	e = xmalloc (sizeof (*e));
 	e->childfd = fdpair[0];
 	e->parentfd = fdpair[1];
@@ -1663,7 +1785,6 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 	int i;
 	struct priv_state sprivs;
 	jobacct_id_t jobacct_id;
-	char *oom_value;
 	List exec_wait_list = NULL;
 	uint32_t jobid;
 	uint32_t node_offset = 0, task_offset = 0;
@@ -1678,7 +1799,6 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 
 	xassert(job != NULL);
 
-	set_oom_adj(0);	/* the tasks may be killed by OOM */
 	if (task_g_pre_setuid(job)) {
 		error("Failed to invoke task plugins: one of task_p_pre_setuid functions returned error");
 		return SLURM_ERROR;
@@ -1694,7 +1814,7 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 	 * Temporarily drop effective privileges, except for the euid.
 	 * We need to wait until after pam_setup() to drop euid.
 	 */
-	if (_drop_privileges (job, false, &sprivs, true) < 0)
+	if (drop_privileges (job, false, &sprivs, true) < 0)
 		return ESLURMD_SET_UID_OR_GID_ERROR;
 
 	if (pam_setup(job->user_name, conf->hostname)
@@ -1706,15 +1826,12 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 	/*
 	 * Reclaim privileges to do the io setup
 	 */
-	_reclaim_privileges (&sprivs);
+	reclaim_privileges(&sprivs);
 	if (rc)
 		goto fail1; /* pam_setup error */
 
 	set_umask(job);		/* set umask for stdout/err files */
-	if (job->flags & LAUNCH_USER_MANAGED_IO)
-		rc = _setup_user_managed_io(job);
-	else
-		rc = _setup_normal_io(job);
+	rc = _setup_normal_io(job);
 	/*
 	 * Initialize log facility to copy errors back to srun
 	 */
@@ -1722,9 +1839,13 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 		rc = _slurmd_job_log_init(job);
 
 	if (rc) {
-		error("IO setup failed: %m");
-		job->task[0]->estatus = 0x0100;
-		step_complete.step_rc = 0x0100;
+		error("%s: IO setup failed: %s", __func__, slurm_strerror(rc));
+
+		job->task[0]->estatus = rc;
+		slurm_mutex_lock(&step_complete.lock);
+		step_complete.step_rc = rc;
+		slurm_mutex_unlock(&step_complete.lock);
+
 		if (job->batch)
 			rc = SLURM_SUCCESS;	/* drains node otherwise */
 		goto fail1;
@@ -1741,9 +1862,9 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 	if (!job->batch && (job->step_id.step_id != SLURM_INTERACTIVE_STEP)
 	    && job->tres_freq) {
 		if (getuid() == (uid_t) 0) {
-			gres_plugin_step_hardware_init(job->step_gres_list,
-						       job->nodeid,
-						       job->tres_freq);
+			gres_g_step_hardware_init(job->step_gres_list,
+						  job->nodeid,
+						  job->tres_freq);
 		} else {
 			error("%s: invalid permissions: cannot initialize GRES hardware unless Slurmd was started as root",
 			      __func__);
@@ -1753,7 +1874,7 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 	/*
 	 * Temporarily drop effective privileges
 	 */
-	if (_drop_privileges (job, true, &sprivs, true) < 0) {
+	if (drop_privileges (job, true, &sprivs, true) < 0) {
 		error ("_drop_privileges: %m");
 		rc = SLURM_ERROR;
 		goto fail2;
@@ -1787,19 +1908,19 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 		struct exec_wait_info *ei;
 
 		acct_gather_profile_g_task_start(i);
-		if ((ei = _fork_child_with_wait_info (i)) == NULL) {
+		if ((ei = _fork_child_with_wait_info(i)) == NULL) {
 			error("child fork: %m");
-			exec_wait_kill_children (exec_wait_list);
+			exec_wait_kill_children(exec_wait_list);
 			rc = SLURM_ERROR;
 			goto fail4;
-		} else if ((pid = _exec_wait_get_pid (ei)) == 0)  { /* child */
+		} else if ((pid = _exec_wait_get_pid(ei)) == 0) { /* child */
 			/*
 			 *  Destroy exec_wait_list in the child.
 			 *   Only exec_wait_info for previous tasks have been
 			 *   added to the list so far, so everything else
 			 *   can be discarded.
 			 */
-			FREE_NULL_LIST (exec_wait_list);
+			FREE_NULL_LIST(exec_wait_list);
 
 			/* jobacctinfo_endpoll();
 			 * closing jobacct files here causes deadlock */
@@ -1811,7 +1932,7 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 			 * Reclaim privileges for the child and call any plugin
 			 * hooks that may require elevated privs
 			 * sprivs.gid_list is already set from the
-			 * _drop_privileges call above, no not reinitialize.
+			 * drop_privileges call above, no not reinitialize.
 			 * NOTE: Only put things in here that are self contained
 			 * and belong in the child.
 			 */
@@ -1835,7 +1956,7 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 			 *   must be called before setpgid() or it is
 			 *   effectively disabled).
 			 */
-			prepare_stdio (job, job->task[i]);
+			prepare_stdio(job, job->task[i]);
 
 			/* Close profiling file descriptors */
 			acct_gather_profile_g_child_forked();
@@ -1846,7 +1967,7 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 			 *   children in any process groups or containers
 			 *   before they make a call to exec(2).
 			 */
-			if (_exec_wait_child_wait_for_parent (ei) < 0)
+			if (_exec_wait_child_wait_for_parent(ei) < 0)
 				_exit(1);
 
 			exec_task(job, i);
@@ -1856,7 +1977,7 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 		 * Parent continues:
 		 */
 
-		list_append (exec_wait_list, ei);
+		list_append(exec_wait_list, ei);
 
 		log_timestamp(time_stamp, sizeof(time_stamp));
 		verbose("task %lu (%lu) started %s",
@@ -1876,18 +1997,12 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 	/*
 	 * Reclaim privileges
 	 */
-	if (_reclaim_privileges (&sprivs) < 0) {
+	if (reclaim_privileges(&sprivs) < 0) {
 		error ("Unable to reclaim privileges");
 		/* Don't bother erroring out here */
 	}
 
-	if ((oom_value = getenv("SLURMSTEPD_OOM_ADJ"))) {
-		int i = atoi(oom_value);
-		debug("Setting slurmstepd oom_adj to %d", i);
-		set_oom_adj(i);
-	}
-
-	if (chdir (sprivs.saved_cwd) < 0) {
+	if (chdir(sprivs.saved_cwd) < 0) {
 		error ("Unable to return to working directory");
 	}
 
@@ -1904,12 +2019,6 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 			      i, job->task[i]->pid, job->pgid);
 		}
 
-		if (task_g_pre_launch_priv(job, job->task[i]->pid) < 0) {
-			error("task_g_pre_launch_priv: %m");
-			rc = SLURM_ERROR;
-			goto fail2;
-		}
-
 		if (proctrack_g_add(job, job->task[i]->pid)
 		    == SLURM_ERROR) {
 			error("proctrack_g_add: %m");
@@ -1929,6 +2038,29 @@ _fork_all_tasks(stepd_step_rec_t *job, bool *io_initialized)
 			jobacct_gather_add_task(job->task[i]->pid, &jobacct_id,
 						0);
 		}
+
+		/*
+		 * Affinity must be set after cgroup is set, or moving pids from
+		 * one cgroup to another will reset affinity.
+		 */
+		if (task_g_pre_set_affinity(job, i) < 0) {
+			error("task_g_pre_set_affinity: %m");
+			rc = SLURM_ERROR;
+			goto fail2;
+		}
+
+		if (task_g_set_affinity(job, i) < 0) {
+			error("task_g_set_affinity: %m");
+			rc = SLURM_ERROR;
+			goto fail2;
+		}
+
+		if (task_g_post_set_affinity(job, i) < 0) {
+			error("task_g_post_set_affinity: %m");
+			rc = SLURM_ERROR;
+			goto fail2;
+		}
+
 		if (spank_task_post_fork (job, i) < 0) {
 			error ("spank task %d post-fork failed", i);
 			rc = SLURM_ERROR;
@@ -1976,7 +2108,7 @@ fail4:
 		error ("Unable to return to working directory");
 	}
 fail3:
-	_reclaim_privileges (&sprivs);
+	reclaim_privileges (&sprivs);
 fail2:
 	FREE_NULL_LIST(exec_wait_list);
 	io_close_task_fds(job);
@@ -2141,6 +2273,7 @@ _wait_for_any_task(stepd_step_rec_t *job, bool waitflag)
 			job->envtp->batch_flag = job->batch;
 			job->envtp->uid = job->uid;
 			job->envtp->user_name = xstrdup(job->user_name);
+			job->envtp->nodeid = job->nodeid;
 
 			/*
 			 * Modify copy of job's environment. Do not alter in
@@ -2156,15 +2289,24 @@ _wait_for_any_task(stepd_step_rec_t *job, bool waitflag)
 
 			setenvf(&job->env, "SLURM_SCRIPT_CONTEXT",
 				"epilog_task");
+			setenvf(&job->env, "SLURMD_NODENAME", "%s",
+				conf->node_name);
+
 			if (job->task_epilog) {
-				_run_script_as_user("user task_epilog",
-						    job->task_epilog,
-						    job, 5, job->env);
+				rc = _run_script_as_user("user task_epilog",
+							 job->task_epilog, job,
+							 5, job->env);
+				if (rc)
+					error("TaskEpilog failed status=%d",
+					      rc);
 			}
 			if (slurm_conf.task_epilog) {
-				_run_script_as_user("slurm task_epilog",
-						    slurm_conf.task_epilog,
-						    job, -1, job->env);
+				rc = _run_script_as_user("slurm task_epilog",
+							 slurm_conf.task_epilog,
+							 job, -1, job->env);
+				if (rc)
+					error("--task-epilog failed status=%d",
+					      rc);
 			}
 
 			if (spank_task_exit (job, t->id) < 0) {
@@ -2305,7 +2447,8 @@ error:
 	return NULL;
 }
 
-static char *_make_batch_script(batch_job_launch_msg_t *msg, char *path)
+static int _make_batch_script(batch_job_launch_msg_t *msg,
+			      stepd_step_rec_t *job)
 {
 	int flags = O_RDWR | O_CREAT | O_EXCL | O_CLOEXEC;
 	int fd, length;
@@ -2314,16 +2457,16 @@ static char *_make_batch_script(batch_job_launch_msg_t *msg, char *path)
 
 	if (msg->script == NULL) {
 		error("%s: called with NULL script", __func__);
-		return NULL;
+		return SLURM_ERROR;
 	}
 
 	/* note: should replace this with a length as part of msg */
 	if ((length = strlen(msg->script)) < 1) {
 		error("%s: called with empty script", __func__);
-		return NULL;
+		return SLURM_ERROR;
 	}
 
-	xstrfmtcat(script, "%s/%s", path, "slurm_script");
+	script = _batch_script_path(job);
 
 	if ((fd = open(script, flags, S_IRWXU)) < 0) {
 		error("couldn't open `%s': %m", script);
@@ -2351,17 +2494,17 @@ static char *_make_batch_script(batch_job_launch_msg_t *msg, char *path)
 	munmap(output, length);
 
 	if (chown(script, (uid_t) msg->uid, (gid_t) -1) < 0) {
-		error("chown(%s): %m", path);
+		error("chown(%s): %m", script);
 		goto error;
 	}
 
-	return script;
+	job->argv[0] = script;
+	return SLURM_SUCCESS;
 
 error:
 	(void) unlink(script);
 	xfree(script);
-	return NULL;
-
+	return SLURM_ERROR;
 }
 
 extern int stepd_drain_node(char *reason)
@@ -2373,7 +2516,6 @@ extern int stepd_drain_node(char *reason)
 	update_node_msg.node_names = conf->node_name;
 	update_node_msg.node_state = NODE_STATE_DRAIN;
 	update_node_msg.reason = reason;
-	update_node_msg.reason_uid = getuid();
 	update_node_msg.weight = NO_VAL;
 	slurm_msg_t_init(&req_msg);
 	req_msg.msg_type = REQUEST_UPDATE_NODE;
@@ -2386,8 +2528,8 @@ extern int stepd_drain_node(char *reason)
 }
 
 static void
-_send_launch_failure(launch_tasks_request_msg_t *msg, slurm_addr_t *cli,
-		     uid_t cli_uid, int rc, uint16_t protocol_version)
+_send_launch_failure(launch_tasks_request_msg_t *msg, slurm_addr_t *cli, int rc,
+		     uint16_t protocol_version)
 {
 	slurm_msg_t resp_msg;
 	launch_tasks_response_msg_t resp;
@@ -2422,7 +2564,7 @@ _send_launch_failure(launch_tasks_request_msg_t *msg, slurm_addr_t *cli,
 	resp_msg.data = &resp;
 	resp_msg.msg_type = RESPONSE_LAUNCH_TASKS;
 	resp_msg.protocol_version = protocol_version;
-	slurm_msg_set_r_uid(&resp_msg, cli_uid);
+	slurm_msg_set_r_uid(&resp_msg, msg->uid);
 
 	memcpy(&resp.step_id, &msg->step_id, sizeof(resp.step_id));
 
@@ -2503,8 +2645,8 @@ _send_complete_batch_script_msg(stepd_step_rec_t *job, int err, int status)
 	req_msg.msg_type= REQUEST_COMPLETE_BATCH_SCRIPT;
 	req_msg.data	= &req;
 
-	info("sending REQUEST_COMPLETE_BATCH_SCRIPT, error:%d status:%d",
-	     err, status);
+	log_flag(PROTOCOL, "sending REQUEST_COMPLETE_BATCH_SCRIPT slurm_rc:%s job_rc:%d",
+		 slurm_strerror(err), status);
 
 	/* Note: these log messages don't go to slurmd.log from here */
 	for (i = 0; i <= MAX_RETRY; i++) {
@@ -2531,9 +2673,8 @@ _send_complete_batch_script_msg(stepd_step_rec_t *job, int err, int status)
 /* If get_list is false make sure ps->gid_list is initialized before
  * hand to prevent xfree.
  */
-static int
-_drop_privileges(stepd_step_rec_t *job, bool do_setuid,
-		 struct priv_state *ps, bool get_list)
+extern int drop_privileges(stepd_step_rec_t *job, bool do_setuid,
+			   struct priv_state *ps, bool get_list)
 {
 	ps->saved_uid = getuid();
 	ps->saved_gid = getgid();
@@ -2571,7 +2712,7 @@ _drop_privileges(stepd_step_rec_t *job, bool do_setuid,
 	}
 
 	if (setgroups(job->ngids, job->gids) < 0) {
-		error("getgroups: %m");
+		error("setgroups: %m");
 		return -1;
 	}
 
@@ -2583,8 +2724,7 @@ _drop_privileges(stepd_step_rec_t *job, bool do_setuid,
 	return SLURM_SUCCESS;
 }
 
-static int
-_reclaim_privileges(struct priv_state *ps)
+extern int reclaim_privileges(struct priv_state *ps)
 {
 	int rc = SLURM_SUCCESS;
 
@@ -2646,7 +2786,7 @@ _slurmd_job_log_init(stepd_step_rec_t *job)
 	/*
 	 *  Connect slurmd stderr to stderr of job
 	 */
-	if ((job->flags & LAUNCH_USER_MANAGED_IO) || (job->flags & LAUNCH_PTY))
+	if (job->flags & LAUNCH_PTY)
 		fd_set_nonblocking(STDERR_FILENO);
 	if (job->task != NULL) {
 		if (dup2(job->task[0]->stderr_fd, STDERR_FILENO) < 0) {
@@ -2655,8 +2795,11 @@ _slurmd_job_log_init(stepd_step_rec_t *job)
 		}
 	}
 
-	verbose("debug level is '%s'",
-		log_num2string(conf->log_opts.stderr_level));
+	verbose("debug levels are stderr='%s', logfile='%s', syslog='%s'",
+		log_num2string(conf->log_opts.stderr_level),
+		log_num2string(conf->log_opts.logfile_level),
+		log_num2string(conf->log_opts.syslog_level));
+
 	return SLURM_SUCCESS;
 }
 
@@ -2822,14 +2965,28 @@ _run_script_as_user(const char *name, const char *path, stepd_step_rec_t *job,
 		   to the container in the parent of the fork.
 		*/
 		if ((jobid != 0) &&	/* Ignore system processes */
+		    !(job->flags & LAUNCH_NO_ALLOC) &&
 		    (container_g_join(jobid, job->uid) != SLURM_SUCCESS))
 			error("container_g_join(%u): %m", job->step_id.job_id);
 
 		argv[0] = (char *)xstrdup(path);
 		argv[1] = NULL;
 
+#ifdef WITH_SELINUX
+		if (setexeccon(job->selinux_context)) {
+			error("Failed to set SELinux context to %s: %m",
+			      job->selinux_context);
+			_exit(1);
+		}
+#else
+		if (job->selinux_context) {
+			error("Built without SELinux support but context was specified");
+			_exit(1);
+		}
+#endif
+
 		sprivs.gid_list = NULL;	/* initialize to prevent xfree */
-		if (_drop_privileges(job, true, &sprivs, false) < 0) {
+		if (drop_privileges(job, true, &sprivs, false) < 0) {
 			error("run_script_as_user _drop_privileges: %m");
 			/* child process, should not return */
 			exit(127);
@@ -2878,7 +3035,7 @@ _run_script_as_user(const char *name, const char *path, stepd_step_rec_t *job,
 		if (rc < 0) {
 			if (errno == EINTR)
 				continue;
-			error("waidpid: %m");
+			error("waitpid: %m");
 			status = 0;
 			break;
 		} else if (rc == 0) {
