@@ -44,52 +44,61 @@
 #include <stdlib.h>
 #include <sys/mman.h>
 #include <unistd.h>
+#include <poll.h>
 
 #include "src/common/assoc_mgr.h"
 #include "src/common/cpu_frequency.h"
 #include "src/common/gres.h"
-#include "src/common/node_select.h"
+#include "src/common/hash.h"
 #include "src/common/plugstack.h"
+#include "src/common/run_command.h"
+#include "src/common/select.h"
+#include "src/common/setproctitle.h"
 #include "src/common/slurm_auth.h"
 #include "src/common/slurm_jobacct_gather.h"
 #include "src/common/slurm_acct_gather_profile.h"
 #include "src/common/slurm_mpi.h"
 #include "src/common/slurm_protocol_api.h"
+#include "src/common/slurm_protocol_pack.h"
 #include "src/common/slurm_rlimits_info.h"
 #include "src/common/stepd_api.h"
 #include "src/common/switch.h"
-#include "src/common/xcgroup_read_config.h"
 #include "src/common/xmalloc.h"
 #include "src/common/xsignal.h"
 #include "src/common/xstring.h"
+#include "src/common/cgroup.h"
 
 #include "src/slurmd/common/core_spec_plugin.h"
+#include "src/slurmd/common/job_container_plugin.h"
+#include "src/slurmd/common/set_oomadj.h"
 #include "src/slurmd/common/slurmstepd_init.h"
-#include "src/slurmd/common/setproctitle.h"
 #include "src/common/slurm_acct_gather_energy.h"
 #include "src/slurmd/common/proctrack.h"
 #include "src/slurmd/common/xcpuinfo.h"
 #include "src/slurmd/slurmd/slurmd.h"
+#include "src/slurmd/slurmstepd/container.h"
 #include "src/slurmd/slurmstepd/mgr.h"
 #include "src/slurmd/slurmstepd/req.h"
 #include "src/slurmd/slurmstepd/slurmstepd.h"
 #include "src/slurmd/slurmstepd/slurmstepd_job.h"
 
-static int _init_from_slurmd(int sock, char **argv,
-			     slurm_addr_t **_cli, uid_t *_cli_uid,
+static int _init_from_slurmd(int sock, char **argv, slurm_addr_t **_cli,
 			     slurm_addr_t **_self, slurm_msg_t **_msg);
 
 static void _dump_user_env(void);
 static void _send_ok_to_slurmd(int sock);
 static void _send_fail_to_slurmd(int sock);
 static void _got_ack_from_slurmd(int);
-static stepd_step_rec_t *_step_setup(slurm_addr_t *cli, uid_t cli_uid,
-				     slurm_addr_t *self, slurm_msg_t *msg);
+static stepd_step_rec_t *_step_setup(slurm_addr_t *cli, slurm_addr_t *self,
+				     slurm_msg_t *msg);
 #ifdef MEMORY_LEAK_DEBUG
 static void _step_cleanup(stepd_step_rec_t *job, slurm_msg_t *msg, int rc);
 #endif
 static int _process_cmdline (int argc, char **argv);
 
+static pthread_mutex_t cleanup_mutex = PTHREAD_MUTEX_INITIALIZER;
+static bool cleanup = false;
+
 /*
  *  List of signals to block in this process
  */
@@ -108,7 +117,6 @@ main (int argc, char **argv)
 {
 	log_options_t lopts = LOG_OPTS_INITIALIZER;
 	slurm_addr_t *cli;
-	uid_t cli_uid;
 	slurm_addr_t *self;
 	slurm_msg_t *msg;
 	stepd_step_rec_t *job;
@@ -118,6 +126,7 @@ main (int argc, char **argv)
 		fatal ("Error in slurmstepd command line");
 
 	slurm_conf_init(NULL);
+	run_command_init();
 
 	xsignal_block(slurmstepd_blocked_signals);
 	conf = xmalloc(sizeof(*conf));
@@ -127,17 +136,19 @@ main (int argc, char **argv)
 
 	log_init(argv[0], lopts, LOG_DAEMON, NULL);
 
-	if (slurm_select_init(1) != SLURM_SUCCESS )
+	if (select_g_init(1) != SLURM_SUCCESS )
 		fatal( "failed to initialize node selection plugin" );
 	if (slurm_auth_init(NULL) != SLURM_SUCCESS)
 		fatal( "failed to initialize authentication plugin" );
+	if (hash_g_init() != SLURM_SUCCESS)
+		fatal("failed to initialize hash plugin");
 
 	/* Receive job parameters from the slurmd */
-	_init_from_slurmd(STDIN_FILENO, argv, &cli, &cli_uid, &self, &msg);
+	_init_from_slurmd(STDIN_FILENO, argv, &cli, &self, &msg);
 
 	/* Create the stepd_step_rec_t, mostly from info in a
 	 * launch_tasks_request_msg_t or a batch_job_launch_msg_t */
-	if (!(job = _step_setup(cli, cli_uid, self, msg))) {
+	if (!(job = _step_setup(cli, self, msg))) {
 		_send_fail_to_slurmd(STDOUT_FILENO);
 		rc = SLURM_ERROR;
 		goto ending;
@@ -190,20 +201,51 @@ extern int stepd_cleanup(slurm_msg_t *msg, stepd_step_rec_t *job,
 			 slurm_addr_t *cli, slurm_addr_t *self,
 			 int rc, bool only_mem)
 {
+	slurm_mutex_lock(&cleanup_mutex);
+
+	if (cleanup)
+		goto done;
+
 	if (!only_mem) {
 		if (job->batch)
 			batch_finish(job, rc); /* sends batch complete message */
 
 		/* signal the message thread to shutdown, and wait for it */
-		eio_signal_shutdown(job->msg_handle);
+		if (job->msg_handle)
+			eio_signal_shutdown(job->msg_handle);
 		pthread_join(job->msgid, NULL);
 	}
 
-	mpi_fini();	/* Remove stale PMI2 sockets */
+	mpi_fini();
+
+	/*
+	 * This call is only done once per step since stepd_cleanup is protected
+	 * agains multiple and concurrent calls.
+	 */
+	proctrack_g_destroy(job->cont_id);
 
 	if (conf->hwloc_xml)
 		(void)remove(conf->hwloc_xml);
 
+	if (job->container)
+		cleanup_container(job);
+
+	run_command_shutdown();
+
+	if (job->step_id.step_id == SLURM_EXTERN_CONT) {
+		uint32_t jobid;
+#ifdef HAVE_NATIVE_CRAY
+		if (job->het_job_id && (job->het_job_id != NO_VAL))
+			jobid = job->het_job_id;
+		else
+			jobid = job->step_id.job_id;
+#else
+		jobid = job->step_id.job_id;
+#endif
+		if (container_g_stepd_delete(jobid))
+			error("container_g_stepd_delete(%u): %m", jobid);
+	}
+
 #ifdef MEMORY_LEAK_DEBUG
 	acct_gather_conf_destroy();
 	(void) core_spec_g_fini();
@@ -211,7 +253,7 @@ extern int stepd_cleanup(slurm_msg_t *msg, stepd_step_rec_t *job,
 
 	fini_setproctitle();
 
-	xcgroup_fini_slurm_cgroup_conf();
+	cgroup_conf_destroy();
 
 	xfree(cli);
 	xfree(self);
@@ -224,8 +266,12 @@ extern int stepd_cleanup(slurm_msg_t *msg, stepd_step_rec_t *job,
 	xfree(conf->node_topo_addr);
 	xfree(conf->node_topo_pattern);
 	xfree(conf->spooldir);
+	xfree(conf->cpu_spec_list);
 	xfree(conf);
 #endif
+	cleanup = true;
+done:
+	slurm_mutex_unlock(&cleanup_mutex);
 	info("done with job");
 	return rc;
 }
@@ -250,7 +296,7 @@ static slurmd_conf_t *read_slurmd_conf_lite(int fd)
 {
 	int rc;
 	int len;
-	Buf buffer = NULL;
+	buf_t *buffer = NULL;
 	slurmd_conf_t *confl, *local_conf = NULL;
 	int tmp_int = 0;
 	List tmp_list = NULL;
@@ -285,30 +331,30 @@ static slurmd_conf_t *read_slurmd_conf_lite(int fd)
 	free_buf(buffer);
 
 	confl->log_opts.prefix_level = 1;
-	confl->log_opts.stderr_level = confl->debug_level;
 	confl->log_opts.logfile_level = confl->debug_level;
-	confl->log_opts.syslog_level = confl->debug_level;
-	/*
-	 * If daemonizing, turn off stderr logging -- also, if
-	 * logging to a file, turn off syslog.
-	 *
-	 * Otherwise, if remaining in foreground, turn off logging
-	 * to syslog (but keep logfile level)
-	 */
-	if (confl->daemonize) {
+
+	if (confl->daemonize)
 		confl->log_opts.stderr_level = LOG_LEVEL_QUIET;
-		if (confl->logfile)
-			confl->log_opts.syslog_level = LOG_LEVEL_QUIET;
+	else
+		confl->log_opts.stderr_level = confl->debug_level;
+
+	if (confl->syslog_debug != LOG_LEVEL_END) {
+		confl->log_opts.syslog_level = confl->syslog_debug;
+	} else if (!confl->daemonize) {
+		confl->log_opts.syslog_level = LOG_LEVEL_QUIET;
+	} else if ((confl->debug_level > LOG_LEVEL_QUIET) && !confl->logfile) {
+		confl->log_opts.syslog_level = confl->debug_level;
 	} else
-		confl->log_opts.syslog_level  = LOG_LEVEL_QUIET;
+		confl->log_opts.syslog_level = LOG_LEVEL_FATAL;
 
 	/*
 	 * LOGGING BEFORE THIS WILL NOT WORK!  Only afterwards will it show
 	 * up in the log.
 	 */
-	log_alter(confl->log_opts, 0, confl->logfile);
+	log_alter(confl->log_opts, SYSLOG_FACILITY_DAEMON, confl->logfile);
 	log_set_timefmt(slurm_conf.log_fmt);
-	debug2("debug level is '%s'.", log_num2string(confl->debug_level));
+	debug2("debug level read from slurmd is '%s'.",
+		log_num2string(confl->debug_level));
 
 	confl->acct_freq_task = NO_VAL16;
 	tmp_int = acct_gather_parse_freq(PROFILE_TASK,
@@ -383,6 +429,7 @@ static int _handle_spank_mode (int argc, char **argv)
 	log_init(prefix, lopts, LOG_DAEMON, NULL);
 	xfree(prefix);
 
+	slurm_conf_init(NULL);
 	/*
 	 *  When we are started from slurmd, a lightweight config is
 	 *   sent over the stdin fd. If we are able to read this conf
@@ -394,8 +441,6 @@ static int _handle_spank_mode (int argc, char **argv)
 	conf = read_slurmd_conf_lite (STDIN_FILENO);
 	close (STDIN_FILENO);
 
-	slurm_conf_init(NULL);
-
 	if (_get_jobid_uid_gid_from_env(&jobid, &uid, &gid))
 		return error("spank environment invalid");
 
@@ -427,6 +472,11 @@ static int _process_cmdline (int argc, char **argv)
 		_dump_user_env();
 		exit(0);
 	}
+	if ((argc == 2) && (xstrcmp(argv[1], "infinity") == 0)) {
+		set_oom_adj(-1000);
+		(void) poll(NULL, 0, -1);
+		exit(0);
+	}
 	if ((argc == 3) && (xstrcmp(argv[1], "spank") == 0)) {
 		if (_handle_spank_mode(argc, argv) < 0)
 			exit (1);
@@ -503,16 +553,14 @@ static void _set_job_log_prefix(slurm_step_id_t *step_id)
  */
 static int
 _init_from_slurmd(int sock, char **argv,
-		  slurm_addr_t **_cli, uid_t *_cli_uid, slurm_addr_t **_self,
-		  slurm_msg_t **_msg)
+		  slurm_addr_t **_cli, slurm_addr_t **_self, slurm_msg_t **_msg)
 {
 	char *incoming_buffer = NULL;
-	Buf buffer;
+	buf_t *buffer;
 	int step_type;
 	int len;
 	uint16_t proto;
 	slurm_addr_t *cli = NULL;
-	uid_t cli_uid;
 	slurm_addr_t *self = NULL;
 	slurm_msg_t *msg = NULL;
 	slurm_step_id_t step_id = {
@@ -526,9 +574,11 @@ _init_from_slurmd(int sock, char **argv,
 		fatal("Failed to read conf from slurmd");
 
 	/* receive cgroup conf from slurmd */
-	if (xcgroup_read_conf(sock) != SLURM_SUCCESS)
+	if (cgroup_read_conf(sock) != SLURM_SUCCESS)
 		fatal("Failed to read cgroup conf from slurmd");
 
+	slurm_conf.slurmd_port = conf->port;
+	setenvf(NULL, "SLURMD_NODENAME", "%s", conf->node_name);
 	/* receive acct_gather conf from slurmd */
 	if (acct_gather_read_conf(sock) != SLURM_SUCCESS)
 		fatal("Failed to read acct_gather conf from slurmd");
@@ -550,8 +600,6 @@ _init_from_slurmd(int sock, char **argv,
 	step_complete.jobacct = jobacctinfo_create(NULL);
 	slurm_mutex_unlock(&step_complete.lock);
 
-	switch_g_slurmd_step_init();
-
 	debug3("slurmstepd rank %d, parent = %pA",
 	       step_complete.rank, &step_complete.parent_addr);
 
@@ -564,7 +612,6 @@ _init_from_slurmd(int sock, char **argv,
 	if (slurm_unpack_addr_no_alloc(cli, buffer) == SLURM_ERROR)
 		fatal("slurmstepd: problem with unpack of slurmd_conf");
 	free_buf(buffer);
-	safe_read(sock, &cli_uid, sizeof(uid_t));
 
 	/* receive self from slurmd */
 	safe_read(sock, &len, sizeof(int));
@@ -634,15 +681,14 @@ _init_from_slurmd(int sock, char **argv,
 	}
 
 	/* Receive GRES information from slurmd */
-	gres_plugin_recv_stepd(sock, msg);
+	gres_g_recv_stepd(sock, msg);
 
-	/*
-	 * Read slurmd node name - may be dictated by slurmctld due to
-	 * Dynamic Future node mapping.
-	 */
-	safe_read(sock, &len, sizeof(int));
-	conf->node_name = xmalloc(len);
-	safe_read(sock, conf->node_name, len);
+	/* Receive mpi.conf from slurmd */
+	if ((step_type == LAUNCH_TASKS) &&
+	    (step_id.step_id != SLURM_EXTERN_CONT) &&
+	    (step_id.step_id != SLURM_INTERACTIVE_STEP) &&
+	    (mpi_conf_recv_stepd(sock) != SLURM_SUCCESS))
+		fatal("Failed to read MPI conf from slurmd");
 
 	_set_job_log_prefix(&step_id);
 
@@ -665,7 +711,6 @@ _init_from_slurmd(int sock, char **argv,
 	msg->protocol_version = proto;
 
 	*_cli = cli;
-	*_cli_uid = cli_uid;
 	*_self = self;
 	*_msg = msg;
 
@@ -677,8 +722,7 @@ rwfail:
 }
 
 static stepd_step_rec_t *
-_step_setup(slurm_addr_t *cli, uid_t cli_uid, slurm_addr_t *self,
-	    slurm_msg_t *msg)
+_step_setup(slurm_addr_t *cli, slurm_addr_t *self, slurm_msg_t *msg)
 {
 	stepd_step_rec_t *job = NULL;
 
@@ -689,7 +733,7 @@ _step_setup(slurm_addr_t *cli, uid_t cli_uid, slurm_addr_t *self,
 		break;
 	case REQUEST_LAUNCH_TASKS:
 		debug2("setup for a launch_task");
-		job = mgr_launch_tasks_setup(msg->data, cli, cli_uid, self,
+		job = mgr_launch_tasks_setup(msg->data, cli, self,
 					     msg->protocol_version);
 		break;
 	default:
@@ -702,22 +746,49 @@ _step_setup(slurm_addr_t *cli, uid_t cli_uid, slurm_addr_t *self,
 		return NULL;
 	}
 
+	if (job->container) {
+	        struct priv_state sprivs;
+		int rc;
+
+		if (drop_privileges(job, false, &sprivs, true) < 0) {
+			error("%s: drop_priviledges failed", __func__);
+			return NULL;
+		}
+		rc = setup_container(job);
+		if (reclaim_privileges(&sprivs) < 0) {
+			error("%s: reclaim_priviledges failed", __func__);
+			return NULL;
+		}
+
+		if (rc == ESLURM_CONTAINER_NOT_CONFIGURED) {
+			debug2("%s: container %s requested but containers are not configured on this node",
+			       __func__, job->container);
+		} else if (rc) {
+			error("%s: container setup failed: %s",
+			      __func__, slurm_strerror(rc));
+			stepd_step_rec_destroy(job);
+			return NULL;
+		} else {
+			debug2("%s: container %s successfully setup",
+			       __func__, job->container);
+		}
+	}
+
 	job->jmgr_pid = getpid();
 	job->jobacct = jobacctinfo_create(NULL);
 
 	/* Establish GRES environment variables */
 	if (slurm_conf.debug_flags & DEBUG_FLAG_GRES) {
-		gres_plugin_job_state_log(job->job_gres_list,
-					  job->step_id.job_id);
-		gres_plugin_step_state_log(job->step_gres_list,
-					   job->step_id.job_id,
-					   job->step_id.step_id);
+		gres_job_state_log(job->job_gres_list,
+				   job->step_id.job_id);
+		gres_step_state_log(job->step_gres_list,
+				    job->step_id.job_id,
+				    job->step_id.step_id);
 	}
 	if (job->batch || (job->step_id.step_id == SLURM_INTERACTIVE_STEP)) {
-		gres_plugin_job_set_env(&job->env, job->job_gres_list, 0);
+		gres_g_job_set_env(&job->env, job->job_gres_list, 0);
 	} else if (msg->msg_type == REQUEST_LAUNCH_TASKS) {
-		gres_plugin_step_set_env(&job->env, job->step_gres_list, 0,
-					 NULL, -1);
+		gres_g_step_set_env(&job->env, job->step_gres_list);
 	}
 
 	/*
@@ -727,6 +798,15 @@ _step_setup(slurm_addr_t *cli, uid_t cli_uid, slurm_addr_t *self,
 			    conf->node_topo_addr);
 	env_array_overwrite(&job->env,"SLURM_TOPOLOGY_ADDR_PATTERN",
 			    conf->node_topo_pattern);
+	/*
+	 * Reset address for cloud nodes
+	 */
+	if (job->alias_list && set_nodes_alias(job->alias_list)) {
+		error("%s: set_nodes_alias failed: %s", __func__,
+		      job->alias_list);
+		stepd_step_rec_destroy(job);
+		return NULL;
+	}
 
 	set_msg_node_id(job);
 
